#!/usr/bin/env python3
# License: BSD-3-Clause

import gzip
import io
from base64 import b64decode
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from astropy import units as u
from astropy import visualization
from astropy.io import fits
from astropy.time import Time
from matplotlib.colors import Normalize
from matplotlib.ticker import MultipleLocator
from ztfquery.utils.stamps import get_ps_stamp

from nuztf.ampel_api import create_empty_cutout, ensure_cutouts
from nuztf.cat_match import get_cross_match_info
from nuztf.paths import CUTOUT_CACHE_DIR
from nuztf.utils import cosmo


def alert_to_pandas(alert):
    candidate = alert[0]["candidate"]
    prv_candid = alert[0]["prv_candidates"]
    combined = [candidate]
    combined.extend(prv_candid)

    df_detections_list = []
    df_ulims_list = []

    for cand in combined:
        _df = pd.DataFrame().from_dict(cand, orient="index").transpose()
        _df["mjd"] = _df["jd"] - 2400000.5
        if "magpsf" in cand.keys() and "isdiffpos" in cand.keys():
            df_detections_list.append(_df)

        else:
            df_ulims_list.append(_df)

    df_detections = pd.concat(df_detections_list)
    if len(df_ulims_list) > 0:
        df_ulims = pd.concat(df_ulims_list)
    else:
        df_ulims = None

    return df_detections, df_ulims


def lightcurve_from_alert(
    alert: list,
    figsize: list = [8, 5],
    title: str = None,
    include_ulims: bool = True,
    include_cutouts: bool = True,
    include_ps1: bool = True,
    include_crossmatch: bool = True,
    mag_range: list = None,
    z: float = None,
    legend: bool = False,
    grid_interval: int = None,
    t_0_mjd: float = None,
    logger=None,
):
    """plot AMPEL alerts as lightcurve"""

    if logger is None:
        import logging

        logger = logging.getLogger(__name__)
    else:
        logger = logger

    if z is not None:
        if np.isnan(z):
            z = None
            logger.debug("Redshift is nan, will be ignored")

    # ZTF color and naming scheme
    BAND_NAMES = {1: "ZTF g", 2: "ZTF r", 3: "ZTF i"}
    BAND_COLORS = {1: "green", 2: "red", 3: "orange"}

    name = alert[0]["objectId"]
    candidate = alert[0]["candidate"]

    if include_cutouts:
        if "cutoutScience" in alert[0].keys():
            if "stampData" in alert[0]["cutoutScience"].keys():
                logger.debug(f"{name}: Cutouts are present.")
            else:
                logger.debug(f"{name}: Cutouts are missing data. Will obtain them")
                alert = ensure_cutouts(alert, logger=logger)
        else:
            logger.debug(
                "The alert dictionary does not contain cutouts. Will obtain them."
            )
            alert = ensure_cutouts(alert, logger=logger)

    logger.debug(f"Plotting {name}")

    df, df_ulims = alert_to_pandas(alert)

    fig = plt.figure(figsize=figsize)

    if include_cutouts:
        lc_ax1 = fig.add_subplot(5, 4, (9, 19))
        cutoutsci = fig.add_subplot(5, 4, (1, 5))
        cutouttemp = fig.add_subplot(5, 4, (2, 6))
        cutoutdiff = fig.add_subplot(5, 4, (3, 7))
        cutoutps1 = fig.add_subplot(5, 4, (4, 8))
    else:
        lc_ax1 = fig.add_subplot(1, 1, 1)
        fig.subplots_adjust(top=0.8, bottom=0.15)

    plt.subplots_adjust(wspace=0.4, hspace=1.8)

    if include_cutouts:
        for cutout_, ax_, type_ in zip(
            [alert[0], alert[0], alert[0]],
            [cutoutsci, cutouttemp, cutoutdiff],
            ["Science", "Template", "Difference"],
        ):
            create_stamp_plot(alert=cutout_, ax=ax_, cutout_type=type_)

        if include_ps1:
            img_cache = CUTOUT_CACHE_DIR.joinpath(f"{name}_PS1.png")

            if not img_cache.is_file():
                img = get_ps_stamp(
                    candidate["ra"], candidate["dec"], size=240, color=["y", "g", "i"]
                )
                img.save(img_cache)

            else:
                from PIL import Image

                img = Image.open(img_cache)

            cutoutps1.imshow(np.asarray(img))
            cutoutps1.set_title("PS1", fontdict={"fontsize": "small"})
            cutoutps1.set_xticks([])
            cutoutps1.set_yticks([])

    # If redshift is given, calculate absolute magnitude via luminosity distance
    # and plot as right axis
    if z is not None:
        dist_l = cosmo.luminosity_distance(z).to(u.pc).value

        def mag_to_absmag(mag):
            absmag = mag - 5 * (np.log10(dist_l) - 1)
            return absmag

        def absmag_to_mag(absmag):
            mag = absmag + 5 * (np.log10(dist_l) - 1)
            return mag

        lc_ax3 = lc_ax1.secondary_yaxis(
            "right", functions=(mag_to_absmag, absmag_to_mag)
        )

        if not include_cutouts:
            lc_ax3.set_ylabel(f"Absolute Magnitude [AB]")

    # Give the figure a title
    if not include_cutouts:
        if title is None:
            fig.suptitle(f"{name}", fontweight="bold")
        else:
            fig.suptitle(title, fontweight="bold")

    if grid_interval is not None:
        lc_ax1.xaxis.set_major_locator(MultipleLocator(grid_interval))

    lc_ax1.grid(visible=True, axis="both", alpha=0.5)
    lc_ax1.set_ylabel("Magnitude [AB]")

    if not include_cutouts:
        lc_ax1.set_xlabel("MJD")

    # Determine magnitude limits
    if mag_range is None:
        max_mag = np.max(df.magpsf.values) + 0.3
        min_mag = np.min(df.magpsf.values) - 0.3
        lc_ax1.set_ylim([max_mag, min_mag])
    else:
        lc_ax1.set_ylim([np.max(mag_range), np.min(mag_range)])

    for fid in BAND_NAMES.keys():
        # Plot older datapoints
        df_temp = df.iloc[1:].query("fid == @fid")
        lc_ax1.errorbar(
            df_temp["mjd"],
            df_temp["magpsf"],
            df_temp["sigmapsf"],
            color=BAND_COLORS[fid],
            fmt=".",
            label=BAND_NAMES[fid],
            mec="black",
            mew=0.5,
        )

        # Plot upper limits
        if df_ulims is not None:
            if include_ulims:
                df_temp2 = df_ulims.query("fid == @fid")
                lc_ax1.scatter(
                    df_temp2["mjd"],
                    df_temp2["diffmaglim"],
                    c=BAND_COLORS[fid],
                    marker="v",
                    s=1.3,
                    alpha=0.5,
                )

    # Plot datapoint from alert
    df_temp = df.iloc[0]
    fid = df_temp["fid"]
    lc_ax1.errorbar(
        df_temp["mjd"],
        df_temp["magpsf"],
        df_temp["sigmapsf"],
        color=BAND_COLORS[fid],
        fmt=".",
        label=BAND_NAMES[fid],
        mec="black",
        mew=0.5,
        markersize=12,
    )

    if legend:
        plt.legend()

    # Now we create an infobox
    if include_cutouts:
        info = []

        info.append(name)
        info.append("------------------------")
        info.append(f"RA: {candidate['ra']:.8f}")
        info.append(f"Dec: {candidate['dec']:.8f}")
        if "drb" in candidate.keys():
            info.append(f"drb: {candidate['drb']:.3f}")
        else:
            info.append(f"rb: {candidate['rb']:.3f}")
        info.append("------------------------")

        for entry in ["sgscore1", "distpsnr1", "srmag1"]:
            info.append(f"{entry[:-1]}: {candidate[entry]:.3f}")

        if alert[0].get("kilonova_eval") is not None:
            info.append(
                f"------------------------\nAMPEL KN score: {alert[0]['kilonova_eval']['kilonovaness']}"
            )

        if (redshift := alert[0].get("redshifts", {}).get("ampel_z")) is not None:
            if alert[0]["redshifts"]["group_z_nbr"] in [1, 2]:
                info.append(f"spec z: {redshift:.3f}")
            else:
                info.append(f"photo z: {redshift:.3f}")

        fig.text(0.77, 0.55, "\n".join(info), va="top", fontsize="medium", alpha=0.5)

    # Add annotations

    lc_ax1.annotate(
        "See On Fritz",
        xy=(0.5, 1),
        xytext=(0.78, 0.10),
        xycoords="figure fraction",
        verticalalignment="top",
        color="royalblue",
        url=f"https://fritz.science/source/{name}",
        fontsize=12,
        bbox=dict(boxstyle="round", fc="cornflowerblue", ec="royalblue", alpha=0.4),
    )

    if include_crossmatch:
        xmatch_info = get_cross_match_info(
            raw=alert[0],
        )
        if include_cutouts:
            ypos = 0.975
        else:
            ypos = 0.035

        if "[TNS NAME=" in xmatch_info:
            tns_name = (
                xmatch_info.split("[TNS NAME=")[1].split("]")[0].strip("AT").strip("SN")
            )
            lc_ax1.annotate(
                "See On TNS",
                xy=(0.5, 1),
                xytext=(0.78, 0.05),
                xycoords="figure fraction",
                verticalalignment="top",
                color="royalblue",
                url=f"https://www.wis-tns.org/object/{tns_name}",
                fontsize=12,
                bbox=dict(
                    boxstyle="round", fc="cornflowerblue", ec="royalblue", alpha=0.4
                ),
            )

        fig.text(
            0.5,
            ypos,
            xmatch_info,
            va="top",
            ha="center",
            fontsize="medium",
            alpha=0.5,
        )

    if t_0_mjd is not None:
        lc_ax1.axvline(t_0_mjd, linestyle=":")
    else:
        t_0_mjd = np.mean(df.mjd.values)

    # Ugly hack because secondary_axis does not work with astropy.time.Time datetime conversion
    mjd_min = min(np.min(df.mjd.values), t_0_mjd)
    mjd_max = max(np.max(df.mjd.values), t_0_mjd)
    length = mjd_max - mjd_min

    lc_ax1.set_xlim([mjd_min - (length / 20), mjd_max + (length / 20)])

    lc_ax2 = lc_ax1.twiny()

    datetimes = [Time(x, format="mjd").datetime for x in [mjd_min, mjd_max]]

    lc_ax2.scatter(
        [Time(x, format="mjd").datetime for x in [mjd_min, mjd_max]], [20, 20], alpha=0
    )
    lc_ax2.tick_params(axis="both", which="major", labelsize=6, rotation=45)
    lc_ax1.tick_params(axis="x", which="major", labelsize=6, rotation=45)
    lc_ax1.ticklabel_format(axis="x", style="plain")
    lc_ax1.tick_params(axis="y", which="major", labelsize=9)

    if z is not None:
        lc_ax3.tick_params(axis="both", which="major", labelsize=9)

    if z is not None:
        axes = [lc_ax1, lc_ax2, lc_ax3]
    else:
        axes = [lc_ax1, lc_ax2]

    return fig, axes


def create_stamp_plot(alert: dict, ax, cutout_type: str):
    """Helper function to create cutout subplot"""
    v3_cutout_names = {
        "Science": "Cutoutscience",
        "Template": "Cutouttemplate",
        "Difference": "Cutoutdifference",
    }

    if alert.get(f"cutout{cutout_type}") is None:
        v3_cutout_type = v3_cutout_names[cutout_type]
        _data = alert.get(f"cutout{v3_cutout_type}", {}).get("stampData", {})
        if _data is not None:
            data = _data.get("stampData")
        else:
            data = None
        if data is None:
            data = create_empty_cutout()
    else:
        data = alert[f"cutout{cutout_type}"]["stampData"]

    with gzip.open(io.BytesIO(b64decode(data)), "rb") as f:
        data = fits.open(io.BytesIO(f.read()), ignore_missing_simple=True)[0].data
    vmin, vmax = np.percentile(data[data == data], [0, 100])
    data_ = visualization.AsinhStretch()((data - vmin) / (vmax - vmin))
    ax.imshow(
        data_,
        norm=Normalize(*np.percentile(data_[data_ == data_], [0.5, 99.5])),
        aspect="auto",
    )
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_title(cutout_type, fontdict={"fontsize": "small"})



#!/usr/bin/env python3
# coding: utf-8

import json
import logging
import time
from pathlib import Path

import backoff
import healpy as hp
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import requests
from ampel.ztf.alert.ZiAlertSupplier import ZiAlertSupplier
from ampel.ztf.dev.DevAlertConsumer import DevAlertConsumer
from ampel.ztf.t0.DecentFilter import DecentFilter
from astropy import units as u
from astropy.coordinates import Distance, SkyCoord
from astropy.time import Time
from matplotlib.backends.backend_pdf import PdfPages
from tqdm import tqdm
from ztfquery import fields as ztfquery_fields

from nuztf.ampel_api import (
    ampel_api_acknowledge_chunk,
    ampel_api_lightcurve,
    ampel_api_name,
    ampel_api_skymap,
)
from nuztf.cat_match import ampel_api_tns, get_cross_match_info, query_ned_for_z
from nuztf.flatpix import get_flatpix, get_nested_pix
from nuztf.fritz import save_source_to_group
from nuztf.observations import get_obs_summary
from nuztf.paths import BASE_CANDIDATE_DIR, RESULTS_DIR
from nuztf.plot import lightcurve_from_alert
from nuztf.utils import cosmo

DEBUG = False
RATELIMIT_CALLS = 10
RATELIMIT_PERIOD = 1


class BaseScanner:
    default_fritz_group = 1430

    def __init__(
        self,
        run_config,
        t_min,
        resource=None,
        filter_class=DecentFilter,
        cone_nside=64,
        cones_to_scan=None,
        logger=None,
    ):
        self.cone_nside = cone_nside
        self.t_min = t_min
        (
            self.map_coords,
            self.pixel_nos,
            self.nside,
            self.map_probs,
            self.data,
            self.total_pixel_area,
            self.key,
        ) = self.unpack_skymap()

        if not hasattr(self, "prob_threshold"):
            self.prob_threshold = None

        if resource is None:
            resource = {
                "ampel-ztf/catalogmatch": "https://ampel.zeuthen.desy.de/api/catalogmatch/",
            }

        if logger is None:
            self.logger = logging.getLogger(__name__)
        else:
            self.logger = logger

        self.logger.info("AMPEL run config:")
        self.logger.info(run_config)

        lvl = self.logger.level

        if lvl > 10:
            logger_ampel = logging.getLogger("AMPEL_filter")
            logger_ampel.setLevel(logging.WARNING)
        else:
            from ampel.log.AmpelLogger import AmpelLogger

            logger_ampel = AmpelLogger()

        self.ampel_filter_class = filter_class(
            logger=logger_ampel, resource=resource, **run_config
        )
        self.ampel_filter_class.post_init()

        self.dac = DevAlertConsumer(self.ampel_filter_class)

        self.scanned_pixels = []

        if cones_to_scan is None:
            self.cone_ids, self.cone_coords = self.find_cone_coords()
        else:
            self.cone_ids, self.cone_coords = cones_to_scan

        self.cache = dict()
        self.default_t_max = t_min + 10.0

        self.overlap_prob = None
        self.overlap_fields = None
        self.first_obs = None
        self.last_obs = None
        self.n_fields = None
        self.rectangular_area = None
        self.double_extragalactic_area = None

        self.observations = None

        self.pixel_area = hp.pixelfunc.nside2pixarea(self.nside, degrees=True)

        if not hasattr(self, "dist"):
            self.dist = None

    def get_full_name(self) -> str:
        raise NotImplementedError

    def get_name(self) -> str:
        raise NotImplementedError

    def get_output_dir(self) -> Path:
        output_dir = RESULTS_DIR.joinpath(self.get_name())
        output_dir.mkdir(exist_ok=True, parents=True)
        return output_dir

    def get_cache_dir(self) -> Path:
        cache_dir = BASE_CANDIDATE_DIR.joinpath(self.get_name().replace(" ", "_"))
        cache_dir.mkdir(exist_ok=True, parents=True)
        return cache_dir

    def unpack_skymap(self):
        raise NotImplementedError

    @staticmethod
    def get_tiling_line():
        return ""

    @staticmethod
    def get_obs_line():
        raise NotImplementedError

    @staticmethod
    def remove_variability_line():
        raise NotImplementedError

    @staticmethod
    def fid_to_band(fid: int):
        fid_band = {1: "g", 2: "r", 3: "i"}
        return fid_band[fid]

    @staticmethod
    def parse_ztf_filter(fid: int):
        """
        Convert ZTF filter id to string

        :param fid: filter id
        :return: g, r, or i
        """
        return ["g", "r", "i"][fid - 1]

    def get_overlap_line(self):
        """ """
        if (self.overlap_prob is not None) and (
            self.double_extragalactic_area is not None
        ):
            return (
                f"We covered {self.overlap_prob:.1f}% "
                f"({self.double_extragalactic_area:.1f} sq deg) "
                f"of the reported localization region. "
                "This estimate accounts for chip gaps. "
            )
        else:
            self.logger.warning("No overlap line added!")
            return ""

    def filter_ampel(self, res):
        self.logger.debug(f"{res['objectId']}: Running AMPEL filter")

        shaped_alert = ZiAlertSupplier.shape_alert_dict(res, ["FilterTest"])
        filterres = self.ampel_filter_class.process(alert=shaped_alert)
        if filterres:
            return True
        else:
            return False

    @backoff.on_exception(
        backoff.expo,
        requests.exceptions.RequestException,
        max_time=600,
    )
    def add_res_to_cache(self, res):
        for res_alert in res:
            if res_alert["objectId"] not in self.cache.keys():
                self.cache[res_alert["objectId"]] = res_alert
            elif (
                res_alert["candidate"]["jd"]
                > self.cache[res_alert["objectId"]]["candidate"]["jd"]
            ):
                self.cache[res_alert["objectId"]] = res_alert

    def add_to_cache_by_names(self, ztf_ids: list):
        for ztf_id in ztf_ids:
            query_res = ampel_api_name(ztf_id, logger=self.logger)
            self.add_res_to_cache(query_res)

    def check_ampel_filter(self, ztf_name):
        lvl = logging.getLogger().getEffectiveLevel()
        logging.getLogger().setLevel(logging.DEBUG)
        self.logger.info("Set logger level to DEBUG")
        all_query_res = ampel_api_name(ztf_name, logger=self.logger)
        assert len(all_query_res) > 0, f"No results from ampel api for {ztf_name}"
        pipeline_bool = False
        for query_res in all_query_res:
            self.logger.info("Checking filter f (no prv)")
            no_prv_bool = self.filter_f_no_prv(query_res)
            self.logger.info(f"Filter f (np prv): {no_prv_bool}")
            if no_prv_bool:
                self.logger.info("Checking ampel filter")
                bool_ampel = self.filter_ampel(query_res)
                self.logger.info(f"ampel filter: {bool_ampel}")
                if bool_ampel:
                    self.logger.info("Checking filter f (history)")
                    history_bool = self.filter_f_history(query_res)
                    self.logger.info(f"Filter f (history): {history_bool}")
                    if history_bool:
                        pipeline_bool = True
        self.logger.info(f"Setting logger back to {lvl}")
        logging.getLogger().setLevel(lvl)
        return pipeline_bool

    def get_multi_night_summary(self, max_days=None):
        mns = get_obs_summary(self.t_min, max_days=max_days)
        # if mns is None:
        #     mns = get_obs_summary(self.t_min, max_days=max_days)
        return mns

    def query_ampel(
        self,
        t_min=None,
        t_max=None,
    ):
        if t_max is None:
            t_max = self.default_t_max

        if t_min is None:
            t_min = self.t_min

        self.logger.info("Commencing skymap scan")

        self.logger.debug(
            f"API skymap search: nside = {self.cone_nside} / "
            f"# pixels = {len(self.cone_ids)} / "
            f"timespan = {t_max.jd-t_min.jd:.1f} days."
        )

        query_res = []

        resume = True
        chunk_size = 2000
        resume_token = None

        i = 0
        total_chunks = 0
        t0 = time.time()

        while resume:
            res, resume_token, chunk_id, remaining_chunks = ampel_api_skymap(
                pixels=self.cone_ids,
                nside=self.cone_nside,
                t_min_jd=t_min.jd,
                t_max_jd=t_max.jd,
                logger=self.logger,
                chunk_size=chunk_size,
                resume_token=resume_token,
                warn_exceeding_chunk=False,
            )
            query_res.extend(res)

            ampel_api_acknowledge_chunk(resume_token=resume_token, chunk_id=chunk_id)

            if i == 0:
                total_chunks = remaining_chunks + 1
                self.logger.info(f"Total chunks: {total_chunks}")

            if remaining_chunks % 50 == 0 and remaining_chunks != 0:
                t1 = time.time()
                processed_chunks = total_chunks - remaining_chunks
                time_per_chunk = (t1 - t0) / processed_chunks
                remaining_time = time_per_chunk * remaining_chunks
                self.logger.info(
                    f"Remaining chunks: {remaining_chunks}. Estimated time to finish: "
                    f"{remaining_time/60:.0f} min"
                )

            if len(res) < chunk_size:
                resume = False
                self.logger.info("Done.")
            else:
                self.logger.debug(
                    f"Chunk size reached ({chunk_size}), commencing next query."
                )
            i += 1

        self.logger.info(f"Ingested {len(query_res)} alerts. Commencing filtering now.")

        return query_res

    def get_initial_cache_path(self) -> Path:
        """
        Get the path to the initial cache file

        :return: Cache file path
        """
        return self.get_cache_dir().joinpath("initial_stage.json")

    def get_final_cache_path(self) -> Path:
        """
        Get the path to the final cache file

        :return: Cache file path
        """
        return self.get_cache_dir().joinpath("final_stage.json")

    def load_from_cache(self):
        """
        Load the cache from the final cache file

        :return: None
        """
        if not self.get_final_cache_path().exists():
            err = (
                f"Final cache file {self.get_final_cache_path()} does not exist. "
                f"Run scan_area first."
            )
            self.logger.error(err)
            raise FileNotFoundError(err)

        with open(self.get_final_cache_path(), "r") as infile:
            results = json.load(infile)

        self.add_results(results)

    def scan_area(
        self,
        t_min=None,
        t_max=None,
    ):
        """
        Retrieve alerts for the healpix map from AMPEL API,
        filter the candidates and create a summary
        """
        query_res = self.query_ampel(t_min=t_min, t_max=t_max)

        with open(self.get_initial_cache_path(), "w") as outfile:
            json.dump(query_res, outfile)

        ztf_ids_first_stage = []
        for res in tqdm(query_res):
            if self.filter_f_no_prv(res):
                if self.filter_ampel(res):
                    ztf_ids_first_stage.append(res["objectId"])

        ztf_ids_first_stage = list(set(ztf_ids_first_stage))

        self.logger.info(
            f"{len(ztf_ids_first_stage)} candidates survive filtering stage 1"
        )

        self.logger.info(f"Retrieving alert history from AMPEL for filtering stage 2")

        results = self.ampel_object_search(ztf_ids=ztf_ids_first_stage)

        with open(self.get_final_cache_path(), "w") as outfile:
            json.dump(results, outfile)

        self.add_results(results)
        self.create_candidate_summary()

    def add_results(self, results):
        """
        Add the results to the cache and create a summary

        :param results: Results from the AMPEL API
        :return: None
        """
        for res in results:
            self.add_res_to_cache(res)

        self.logger.info(f"Found {len(self.cache)} candidates")

    def filter_f_no_prv(self, res):
        raise NotImplementedError

    def filter_f_history(self, res):
        raise NotImplementedError

    def find_cone_coords(self):
        raise NotImplementedError

    @staticmethod
    def wrap_around_180(ra_deg: float):
        """ """
        ra_rad = np.deg2rad(ra_deg)
        ra_rad[ra_rad > np.pi] -= 2 * np.pi
        ra_deg = np.rad2deg(ra_rad)
        return ra_deg

    def ampel_object_search(self, ztf_ids: list) -> list:
        """ """
        all_results = []

        for ztf_id in tqdm(ztf_ids):
            # get the full lightcurve from the API
            query_res = ampel_api_lightcurve(ztf_name=ztf_id, logger=self.logger)

            final_res = []

            for res in query_res:
                if self.filter_f_history(res):
                    final_res.append(res)

            all_results.append(final_res)

        return all_results

    @staticmethod
    def calculate_abs_mag(mag: float, redshift: float) -> float:
        """ """
        luminosity_distance = cosmo.luminosity_distance(redshift).value * 10**6
        abs_mag = mag - 5 * (np.log10(luminosity_distance) - 1)
        return abs_mag

    def get_candidates_lines(self) -> str:
        """
        Return a string with the candidates in a format that can be
        directly pasted into a Slack message or GCN

        Returns: str
        """

        if len(self.cache) > 0:
            s = (
                "We are left with the following high-significance transient "
                "candidates by our pipeline, all lying within the "
                f"{100 * self.prob_threshold}% localization of the skymap."
                f"\n\n{self.parse_candidates()}"
            )
        else:
            s = "\n\nNo candidate counterparts were detected."

        return s

    def parse_candidates(self):
        """
        Parse the candidates into a string that can be pasted into a Slack message
        or GCN

        :return: Str
        """

        table = (
            "+--------------------------------------------------------------------------------+\n"
            "| ZTF Name     | IAU Name  | RA (deg)    | DEC (deg)   | Filter | Mag   | MagErr |\n"
            "+--------------------------------------------------------------------------------+\n"
        )
        for name, res in sorted(self.cache.items()):
            jds = [x["jd"] for x in res["prv_candidates"]]

            if res["candidate"]["jd"] > max(jds):
                latest = res["candidate"]
            else:
                latest = res["prv_candidates"][jds.index(max(jds))]

            old_flag = ""

            second_det = [x for x in jds if x > min(jds) + 0.01]

            if len(second_det) > 0:
                if Time.now().jd - second_det[0] > 1.0:
                    old_flag = "(MORE THAN ONE DAY SINCE SECOND DETECTION)"

            tns_result = " ------- "
            tns_name, tns_date, tns_group = ampel_api_tns(
                latest["ra"], latest["dec"], searchradius_arcsec=3
            )
            if tns_name:
                tns_result = tns_name

            line = "| {0} | {1} | {2:011.7f} | {3:+011.7f} | {4}      | {5:.2f} | {6:.2f}   | {7} \n".format(
                name,
                tns_result,
                float(latest["ra"]),
                float(latest["dec"]),
                ["g", "r", "i"][latest["fid"] - 1],
                latest["magpsf"],
                latest["sigmapsf"],
                old_flag,
                sign="+",
                prec=7,
            )
            table += line

        table += "+--------------------------------------------------------------------------------+\n\n"
        return table

    def draft_gcn(self):
        if self.first_obs is None:
            raise ValueError("No first observation set. Cannot draft GCN.")

        first_obs_dt = self.first_obs.datetime
        pretty_date = first_obs_dt.strftime("%Y-%m-%d")
        pretty_time = first_obs_dt.strftime("%H:%M")

        text = (
            f"Astronomer Name (Institute of Somewhere), ............. report,\n\n"
            f"On behalf of the Zwicky Transient Facility (ZTF) and Global Relay of Observatories Watching Transients Happen (GROWTH) collaborations: \n\n"
            f"As part of the ZTF neutrino follow up program (Stein et al. 2023), we observed the localization region of the {self.get_full_name()} with the Palomar 48-inch telescope, equipped with the 47 square degree ZTF camera (Bellm et al. 2019, Graham et al. 2019). {self.get_tiling_line()}"
            f"We started observations in the g- and r-band beginning at {pretty_date} {pretty_time} UTC, "
            f"approximately {(self.first_obs.jd - self.t_min.jd) * 24.0:.1f} hours after event time. {self.get_overlap_line()}"
            f"{self.get_obs_line()} \n \n"
            "The images were processed in real-time through the ZTF reduction and image subtraction pipelines at IPAC to search for potential counterparts (Masci et al. 2019). "
            "AMPEL (Nordin et al. 2019, Stein et al. 2021) was used to search the alerts database for candidates. "
            "We reject stellar sources (Tachibana and Miller 2018) and moving objects, and "
            f"apply machine learning algorithms (Mahabal et al. 2019) {self.remove_variability_line()}. "
            f"{self.get_candidates_lines()} \n\n"
        )

        if self.dist:
            text += (
                "The GW distance estimate is {:.0f} [{:.0f} - {:.0f}] Mpc.\n\n".format(
                    self.dist, self.dist - self.dist_unc, self.dist + self.dist_unc
                )
            )
        else:
            pass

        text += self.text_summary()

        text += (
            "ZTF and GROWTH are worldwide collaborations comprising Caltech, USA; IPAC, USA; WIS, Israel; OKC, Sweden; JSI/UMd, USA; DESY, Germany; TANGO, Taiwan; UW Milwaukee, USA; LANL, USA; TCD, Ireland; IN2P3, France.\n\n"
            "GROWTH acknowledges generous support of the NSF under PIRE Grant No 1545949.\n"
            "Alert distribution service provided by DIRAC@UW (Patterson et al. 2019).\n"
            "Alert database searches are done by AMPEL (Nordin et al. 2019).\n"
            "Alert filtering is performed with the nuztf (Stein et al. 2021, https://github.com/desy-multimessenger/nuztf ).\n"
        )

        return text

    @staticmethod
    def extract_ra_dec(nside: int, index: int):
        """
        Convert a healpix index to ra and dec

        :param nside: Nside of healpix map
        :param index: index of healpix pixel
        :return: ra, dec in degrees
        """
        theta, phi = hp.pix2ang(nside, index, nest=True)
        ra_deg = np.rad2deg(phi)
        dec_deg = np.rad2deg(0.5 * np.pi - theta)

        return ra_deg, dec_deg

    @staticmethod
    def extract_npix(nside: int, ra_deg: float, dec_deg: float) -> int:
        """
        Extract healpix index from ra and dec

        :param nside: nside of healpix map
        :param ra_deg: ra in degrees
        :param dec_deg: dec in degrees
        :return: index of healpix pixel
        """
        theta = 0.5 * np.pi - np.deg2rad(dec_deg)
        phi = np.deg2rad(ra_deg)

        return int(hp.ang2pix(nside, theta, phi, nest=True))

    def create_candidate_summary(self, include_ps1: bool = True) -> str:
        """
        Create pdf with lightcurve plots of all candidates

        :return None:
        """
        if len(self.cache.items()) == 0:
            self.logger.info("No candidates found, skipping pdf creation")
            return

        pdf_path = self.get_output_dir() / "candidates.pdf"

        self.logger.info(
            f"Creating overviewpdf\n(this might take a moment as it involves catalog matching)"
        )
        self.logger.debug(f"Overview pdf path: {pdf_path}")

        with PdfPages(pdf_path) as pdf:
            for name, alert in tqdm(sorted(self.cache.items())):
                fig, _ = lightcurve_from_alert(
                    [alert],
                    include_cutouts=True,
                    include_ps1=include_ps1,
                    logger=self.logger,
                    t_0_mjd=self.t_min.mjd,
                )

                pdf.savefig()
                plt.close()

    def create_overview_table(self):
        """
        Create csv table of all candidates

        :return None:
        """
        if len(self.cache.items()) == 0:
            self.logger.info("No candidates found, skipping csv creation")
            return

        csv_path = self.get_output_dir() / "candidate_table.csv"

        self.logger.info("Creating overview csv")
        self.logger.debug(f"Overview csv path: {csv_path}")

        data = {
            "ztf_id": [],
            "RA": [],
            "Dec": [],
            "mag": [],
            "xmatch": [],
            "kilonova_score": [],
        }

        for ztf_id, alert in tqdm(sorted(self.cache.items())):
            data["ztf_id"].append(ztf_id)
            data["RA"].append(alert["candidate"]["ra"])
            data["Dec"].append(alert["candidate"]["dec"])
            data["mag"].append(alert["candidate"]["magpsf"])
            data["xmatch"].append(get_cross_match_info(raw=alert, logger=self.logger))
            if alert.get("kilonova_eval") is not None:
                data["kilonova_score"].append(alert["kilonova_eval"]["kilonovaness"])
            else:
                data["kilonova_score"].append(None)

        df = pd.DataFrame.from_dict(data)

        df.to_csv(csv_path)

    def tns_summary(self):
        """
        Create summary for TNS

        :return:
        """
        summary = ""

        for name, res in sorted(self.cache.items()):
            detections = [
                x
                for x in res["prv_candidates"] + [res["candidate"]]
                if "isdiffpos" in x.keys()
            ]
            detection_jds = [x["jd"] for x in detections]
            first_detection = detections[detection_jds.index(min(detection_jds))]
            latest = [
                x
                for x in res["prv_candidates"] + [res["candidate"]]
                if "isdiffpos" in x.keys()
            ][-1]
            summary += f"Candidate: {name} / RA={res['candidate']['ra']} / Dec={res['candidate']['dec']} / First detection={first_detection['jd']}\n"
            try:
                last_upper_limit = [
                    x
                    for x in res["prv_candidates"]
                    if np.logical_and(
                        "isdiffpos" in x.keys(), x["jd"] < first_detection["jd"]
                    )
                ][-1]
                summary += f"Last Upper Limit: {last_upper_limit['jd']} / band={self.parse_ztf_filter(last_upper_limit['fid'])} / maglim={last_upper_limit['diffmaglim']:.3f}\n"

            except IndexError:
                last_upper_limit = None
                summary += "Last Upper Limit: None\n"

            summary += f"First Detection: {first_detection['jd']} / band={self.parse_ztf_filter(first_detection['fid'])} / mag={first_detection['magpsf']:.3f} +/- {first_detection['sigmapsf']:.3f}\n"

            hours_after_merger = 24.0 * (first_detection["jd"] - self.t_min.jd)

            summary += f"First observed {hours_after_merger:.2f} hours after merger\n"

            if last_upper_limit:
                summary += f"It has risen {-latest['magpsf'] + last_upper_limit['diffmaglim']} / band={self.parse_ztf_filter(latest['fid'])} / Last upper limit was in band {self.parse_ztf_filter(last_upper_limit['fid'])}\n"

            summary += f"{[x['jd'] for x in res['prv_candidates'] + [res['candidate']] if 'isdiffpos' in x.keys()]}\n"

        self.logger.info(summary)

        return summary

    def peak_mag_summary(self):
        """
        Print a summary of each candidate's peak magnitude

        :return: str
        """

        for name, res in sorted(self.cache.items()):
            detections = [
                x
                for x in res["prv_candidates"] + [res["candidate"]]
                if "isdiffpos" in x.keys()
            ]
            detection_mags = [x["magpsf"] for x in detections]
            brightest = detections[detection_mags.index(min(detection_mags))]

            diff = 0.0
            df = None

            for fid in [1, 2, 3]:
                dets = [x["magpsf"] for x in detections if int(x["fid"]) == fid]
                if len(dets) > 1:
                    nd = max(dets) - min(dets)

                    if nd > diff:
                        diff = nd
                        df = self.parse_ztf_filter(fid)

            tns_result = ""
            tns_name, tns_date, tns_group = ampel_api_tns(
                brightest["ra"], brightest["dec"], searchradius_arcsec=3.0
            )
            if tns_name:
                tns_result = f"({tns_name})"

            xmatch_info = get_cross_match_info(raw=res, logger=self.logger)

            print(
                f"Candidate {name} peaked at {brightest['magpsf']:.1f} {tns_result} "
                f"on {brightest['jd']:.1f} with "
                f"filter {self.parse_ztf_filter(brightest['fid'])}. "
                f"Max range of {diff:.1f} mag with filter {df}. {xmatch_info}"
            )

    def candidate_text(
        self, ztf_id: str, first_detection: float, lul_lim: float, lul_jd: float
    ) -> str:
        """
        Format a text summary of a candidate

        :param ztf_id: ZTF name of the candidate
        :param first_detection: first detection of the candidate
        :param lul_lim: last upper limit magnitude
        :param lul_jd: last upper limit julian date
        :return: text summary
        """
        fd = Time(first_detection, format="jd").datetime.strftime("%Y-%m-%d")

        text = f"{ztf_id} was first detected on {fd}. "

        return text

    def text_summary(self):
        """
        Create a text summary of all candidates

        :return:
        """
        text = ""
        for name, res in sorted(self.cache.items()):
            detections = [
                x
                for x in res["prv_candidates"] + [res["candidate"]]
                if "isdiffpos" in x.keys()
            ]
            detection_jds = [x["jd"] for x in detections]
            first_detection = detections[detection_jds.index(min(detection_jds))]

            latest = [
                x
                for x in res["prv_candidates"] + [res["candidate"]]
                if "isdiffpos" in x.keys()
            ][-1]

            upper_limits = [
                x for x in res["prv_candidates"] if x["jd"] < first_detection["jd"]
            ]
            if len(upper_limits) > 0:
                last_upper_limit_mag = upper_limits[-1]["diffmaglim"]
                last_upper_limit_jd = upper_limits[-1]["jd"]
            else:
                last_upper_limit_mag = None
                last_upper_limit_jd = None

            text += self.candidate_text(
                ztf_id=name,
                first_detection=first_detection["jd"],
                lul_lim=last_upper_limit_mag,
                lul_jd=last_upper_limit_jd,
            )

            ned_z, ned_dist = query_ned_for_z(
                ra_deg=latest["ra"],
                dec_deg=latest["dec"],
                searchradius_arcsec=1,
                logger=self.logger,
            )

            if ned_z:
                ned_z = float(ned_z)
                absmag = self.calculate_abs_mag(latest["magpsf"], ned_z)
                if ned_z > 0:
                    z_dist = Distance(z=ned_z, cosmology=cosmo).value
                    text += (
                        f"It has a spec-z of {ned_z:.3f} [{z_dist:.0f} Mpc] "
                        f"and an abs. mag of {absmag:.1f}. "
                        f"Distance to SDSS galaxy is {ned_dist:.2f} arcsec. "
                    )

            c = SkyCoord(res["candidate"]["ra"], res["candidate"]["dec"], unit="deg")
            g_lat = c.galactic.b.degree
            if abs(g_lat) < 15.0:
                text += f"It is located at a galactic latitude of {g_lat:.2f} degrees. "

            xmatch_info = get_cross_match_info(raw=res, logger=self.logger)
            text += xmatch_info
            text += "\n"

        if len(text) > 0:
            text = f"Amongst our candidates, \n\n{text}\n\n"

        return text

    def calculate_overlap_with_observations(
        self, first_det_window_days=3.0, min_sep=0.01
    ):
        mns = get_obs_summary(t_min=self.t_min, max_days=first_det_window_days)

        if mns is None:
            return None, None, None

        data = mns.data.copy()

        mask = data["status"] == 0
        self.logger.info(
            f"Found {mask.sum()} successful observations in the depot, "
            f"corresponding to {np.mean(mask)*100:.2f}% of the total."
        )

        self.logger.info("Unpacking observations")

        if self.nside > 256:
            (
                self.map_coords,
                self.pixel_nos,
                self.nside,
                self.map_probs,
                self.data,
                self.total_pixel_area,
                self.key,
            ) = self.unpack_skymap(output_nside=256)

        pix_map = dict()
        pix_obs_times = dict()

        nested_pix = get_nested_pix(nside=self.nside, logger=self.logger)

        for i, obs_time in enumerate(tqdm(list(set(data["obsjd"])))):
            obs = data[data["obsjd"] == obs_time]

            field = obs["field_id"].iloc[0]

            try:
                flat_pix = nested_pix[int(field)]

                mask = obs["status"].astype(int) == 0
                indices = obs["qid"].values[mask]

                for qid in indices:
                    pixels = flat_pix[int(qid)]

                    for p in pixels:
                        if p not in pix_obs_times.keys():
                            pix_obs_times[p] = [obs_time]
                        else:
                            pix_obs_times[p] += [obs_time]

                        if p not in pix_map.keys():
                            pix_map[p] = [field]
                        else:
                            pix_map[p] += [field]

            except (KeyError, ValueError):
                self.logger.warning(
                    f"Field {field} not found in nested pix dict. "
                    f"This might be an engineering observation."
                )

        npix = hp.nside2npix(self.nside)
        theta, phi = hp.pix2ang(self.nside, np.arange(npix), nest=False)
        radecs = SkyCoord(ra=phi * u.rad, dec=(0.5 * np.pi - theta) * u.rad)

        ras, decs = hp.pixelfunc.pix2ang(
            self.nside, hp.nest2ring(self.nside, self.pixel_nos), lonlat=True
        )

        coverage_data = []

        overlapping_fields = []
        times = []

        for i, p in enumerate(tqdm(hp.nest2ring(self.nside, self.pixel_nos))):
            entry = {
                "pixel_no": p,
                "prob": self.map_probs[i],
                "gal_b": radecs[i].galactic.b.deg,
                "in_plane": abs(radecs[i].galactic.b.deg) < 10.0,
                "ra_deg": ras[i],
                "dec_deg": decs[i],
            }

            if p in pix_obs_times.keys():
                obs = pix_obs_times[p]

                entry["n_det_class"] = 2 if max(obs) - min(obs) > min_sep else 1

                entry["latency_days"] = min(obs) - self.t_min.jd

                overlapping_fields += pix_map[p]

                times += list(obs)
            else:
                entry["n_det_class"] = 0
                entry["latency_days"] = np.nan

            coverage_data.append(entry)

        coverage_df = pd.DataFrame(coverage_data)

        overlapping_fields = sorted(list(set(overlapping_fields)))

        _observations = data.query("obsjd in @times").reset_index(drop=True)[
            ["obsjd", "exposure_time", "filter_id"]
        ]
        bands = [self.fid_to_band(fid) for fid in _observations["filter_id"].values]
        _observations["band"] = bands
        _observations.drop(columns=["filter_id"], inplace=True)
        self.observations = _observations

        self.logger.info("All observations:")
        self.logger.info(f"\n{self.observations}")

        try:
            self.first_obs = Time(min(times), format="jd")
            self.first_obs.utc.format = "isot"
            self.last_obs = Time(max(times), format="jd")
            self.last_obs.utc.format = "isot"

        except ValueError:
            err = (
                f"No observations of this event were found at any time between "
                f"{self.t_min} and {self.t_min + first_det_window_days * u.day}. "
                f"Coverage overlap is 0%!"
            )
            self.logger.error(err)
            raise ValueError(err)

        self.logger.info(f"Observations started at {self.first_obs.isot}")

        return (
            coverage_df,
            times,
            overlapping_fields,
        )

    def plot_overlap_with_observations(self, first_det_window_days=None, min_sep=0.01):
        """
        Function to plot the overlap of the field with observations.

        :param first_det_window_days: Window of time in days to consider for the first detection.
        :param min_sep: Minimum separation between observations to consider them as separate.

        """

        (
            coverage_df,
            times,
            overlapping_fields,
        ) = self.calculate_overlap_with_observations(
            first_det_window_days=first_det_window_days,
            min_sep=min_sep,
        )

        if coverage_df is None:
            self.logger.warning("Not plotting overlap with observations.")
            return

        fig = plt.figure()
        plt.subplot(projection="aitoff")

        self.overlap_fields = list(set(overlapping_fields))

        overlap_mask = (coverage_df["n_det_class"] == 2) & ~coverage_df["in_plane"]
        self.overlap_prob = coverage_df[overlap_mask]["prob"].sum() * 100.0

        size = hp.max_pixrad(self.nside) ** 2 * 50.0

        veto_pixels = coverage_df.where(coverage_df["n_det_class"] == 0)

        if len(veto_pixels) > 0:
            plt.scatter(
                np.radians(veto_pixels["ra_deg"]),
                np.radians(veto_pixels["dec_deg"]),
                color="red",
                s=size,
            )

        plane_pixels = coverage_df.where(
            coverage_df["in_plane"] & (coverage_df["n_det_class"] > 0)
        )

        if len(plane_pixels) > 0:
            plt.scatter(
                np.radians(self.wrap_around_180(plane_pixels["ra_deg"])),
                np.radians(plane_pixels["dec_deg"]),
                color="green",
                s=size,
            )

        single_pixels = coverage_df.where(
            ~coverage_df["in_plane"] & (coverage_df["n_det_class"] == 1)
        )

        if len(single_pixels) > 0:
            plt.scatter(
                np.radians(self.wrap_around_180(single_pixels["ra_deg"])),
                np.radians(single_pixels["dec_deg"]),
                c=single_pixels["prob"],
                vmin=0.0,
                vmax=max(self.data[self.key]),
                s=size,
                cmap="gray",
            )

        double_pixels = coverage_df.where(
            ~coverage_df["in_plane"] & (coverage_df["n_det_class"] == 2)
        )

        if len(double_pixels) > 0:
            plt.scatter(
                np.radians(self.wrap_around_180(double_pixels["ra_deg"])),
                np.radians(double_pixels["dec_deg"]),
                c=double_pixels["prob"],
                vmin=0.0,
                vmax=max(self.data[self.key]),
                s=size,
            )

        red_patch = mpatches.Patch(color="red", label="Not observed")
        gray_patch = mpatches.Patch(color="gray", label="Observed once")
        violet_patch = mpatches.Patch(
            color="green", label="Observed Galactic Plane (|b|<10)"
        )
        plt.legend(handles=[red_patch, gray_patch, violet_patch])

        message = (
            "In total, {0:.1f} % of the contour was observed at least once.\n"
            "This estimate includes {1:.1f} % of the contour "
            "at a galactic latitude <10 deg.\n"
            "In total, {2:.1f} % of the contour was observed at least twice. \n"
            "In total, {3:.1f} % of the contour was observed at least twice, "
            "and excluding low galactic latitudes.\n"
            "These estimates account for chip gaps.".format(
                100.0 * coverage_df.query("n_det_class > 0")["prob"].sum(),
                100.0 * coverage_df.query("in_plane & n_det_class > 0")["prob"].sum(),
                100.0 * coverage_df.query("n_det_class == 2")["prob"].sum(),
                100.0 * coverage_df.query("n_det_class == 2 & ~in_plane")["prob"].sum(),
            )
        )

        n_pixels = len(coverage_df.query("n_det_class > 0 & prob > 0.0"))
        n_double = len(coverage_df.query("n_det_class == 2 & prob > 0.0"))
        n_plane = len(coverage_df.query("in_plane & n_det_class > 0 & prob > 0.0"))

        self.healpix_area = self.pixel_area * n_pixels
        self.double_extragalactic_area = self.pixel_area * n_double
        plane_area = self.pixel_area * n_plane

        self.overlap_fields = overlapping_fields

        self.logger.info(
            f"{n_pixels} pixels were covered, covering approximately "
            f"{self.healpix_area:.2g} sq deg."
        )
        self.logger.info(
            f"{n_double} pixels were covered at least twice (b>10), "
            f"covering approximately {self.double_extragalactic_area:.2g} sq deg."
        )
        self.logger.info(
            f"{n_plane} pixels were covered at low galactic latitude, "
            f"covering approximately {plane_area:.2g} sq deg."
        )
        return fig, message

    def get_exposure_summary(self) -> pd.DataFrame:
        """
        Function to get a summary of the exposures

        :return:
        """

        if self.observations is None:
            raise ValueError("No pre-observations loaded. Run plot_coverage first")

        max_days = max(self.observations["obsjd"]) - self.t_min.jd

        mns = get_obs_summary(t_min=self.t_min, max_days=max_days)

        exposures = []

        for obsjd in self.observations["obsjd"].unique():
            res = mns.data[mns.data["obsjd"] == obsjd]

            f_proc = float(len(res)) / 64.0
            f_status = np.sum(res["status"].astype(int) == 0) / 64.0

            new = {}

            copy_keys = [
                "exposure_id",
                "obsjd",
                "filter_id",
                "field_id",
                "exposure_time",
                "date",
            ]

            for key in copy_keys:
                new[key] = res[key].iloc[0]

            new["f_processed"] = f_proc
            new["f_proc_without_flag"] = f_status
            exposures.append(new)

        return pd.DataFrame(exposures)

    def get_field_summary(self):
        """
        Function to get a summary of the exposures

        :return:
        """

        df = self.get_exposure_summary()

        fields = []

        for field_entry in df["field_id"].unique():
            res = df[df["field_id"] == field_entry]

            mask = res["exposure_time"].astype(float) > 30.0

            new = {
                "field_id": field_entry,
                "n_exposures": len(res),
                "n_deep": np.sum(mask),
                "n_30": np.sum(~mask),
                "f_proc_deep": np.mean(res[mask]["f_processed"]),
                "f_proc_30": np.mean(res[~mask]["f_processed"]),
                "f_proc_without_flag_deep": np.mean(res[mask]["f_proc_without_flag"]),
                "f_proc_without_flag_30": np.mean(res[~mask]["f_proc_without_flag"]),
                "n_filters": len(res["filter_id"].unique()),
                "n_nights": len(set([int(x + 0.5) for x in res["obsjd"]])),
            }
            fields.append(new)

        summary = pd.DataFrame(fields)

        return summary.sort_values(by="field_id", ignore_index=True)

    def export_cache_to_fritz(self, group_id=None):
        if group_id is None:
            group_id = self.default_fritz_group

        saved_sources = []

        for source in self.cache.keys():
            response = save_source_to_group(object_id=source, group_id=group_id)

            if response.status_code not in [200]:
                self.logger.warning(f"Bad API call for source {source}: {response}")
            else:
                saved_sources.append(source)

        self.logger.info(f"Saved {len(saved_sources)} to fritz group {group_id}")



import json
import logging
import os
import time
from glob import glob
from pathlib import Path

import backoff
import numpy as np
import pandas as pd
import pyvo.dal
import requests
from astropy import units as u
from astropy.time import Time
from pyvo.auth import authsession, securitymethods
from requests.auth import HTTPBasicAuth
from requests.exceptions import HTTPError
from tqdm import tqdm
from ztfquery import skyvision
from ztfquery.io import LOCALSOURCE

from nuztf import credentials

coverage_dir = Path(LOCALSOURCE).joinpath("all_obs")
coverage_dir.mkdir(exist_ok=True)

partial_flag = "_PARTIAL"

logger = logging.getLogger(__name__)

# IPAC TAP login
username, password = credentials.load_credentials("ipacdepot")
data = {"username": username, "password": password}
headers = {"Content-Type": "application/x-www-form-urlencoded", "Accept": "text/plain"}
IPAC_TAP_URL = "https://irsa.ipac.caltech.edu"


class MNS:
    def __init__(self, df):
        self.data = df


class NoDepotEntry(Exception):
    """
    No entry in the depot for a given date
    """


def get_date(jd: float) -> str:
    """
    Get the date in YYYYMMDD format from a JD

    :param jd: JD
    :return: String date
    """
    return str(Time(jd, format="jd").isot).split("T")[0].replace("-", "")


def download_depot_log(date):
    """
    Download the depot log for a given date

    :param date: date in YYYYMMDD format
    :return: json log
    """
    url = f"https://ztfweb.ipac.caltech.edu/ztf/depot/{date}/ztf_recentproc_{date}.json"
    response = requests.get(url, auth=HTTPBasicAuth(username, password))
    if response.status_code == 404:
        raise NoDepotEntry(f"No depot entry for {date} at url {url}")
    response.raise_for_status()
    return response.json()


# The following functions are used to find the paths to the cached coverage files


def coverage_depot_path(jd: float) -> Path:
    """
    Return the path to the cached coverage file for a given JD

    :param jd: JD
    :return: path to cached coverage file
    """

    date = get_date(jd)

    if (Time.now().jd - jd) < 1:
        partial_ext = partial_flag
    else:
        partial_ext = ""

    return coverage_dir.joinpath(f"{date}{partial_ext}.json")


def coverage_skyvision_path(jd: float) -> Path:
    """
    Find the path to the Skyvision coverage file for a given JD

    :param jd: JD
    :return: Output path
    """
    if (Time.now().jd - jd) < 1:
        partial_ext = partial_flag
    else:
        partial_ext = ""

    output_path = coverage_dir.joinpath(f"{get_date(jd)}{partial_ext}_skyvision.json")

    return output_path


def coverage_tap_path(jd: float) -> Path:
    """
    Find the path to the TAP coverage file for a given JD

    :param jd: JD
    :return: Output path
    """
    if (Time.now().jd - jd) < 1:
        partial_ext = partial_flag
    else:
        partial_ext = ""

    output_path = coverage_dir.joinpath(f"{get_date(jd)}{partial_ext}_tap.json")

    return output_path


# The following functions are used to write the coverage to the cache


def write_coverage_depot(jds: list[float]):
    """
    Write the depot coverage for a list of JDs to the cache

    Requires a valid IPAC depot login

    :param jds: JDs
    :return: None
    """
    for jd in tqdm(jds):
        date = get_date(jd)
        try:
            log = download_depot_log(date)
        except NoDepotEntry as exc:
            log = {}
            logger.warning(f"No depot entry for {date}: {exc}")

        path = coverage_depot_path(jd)
        with open(path, "w") as f:
            json.dump(log, f)


@backoff.on_exception(backoff.expo, requests.exceptions.RequestException, max_time=60)
def write_coverage_tap(jds: [float]):
    """
    Write the coverage for a list of JDs to the cache using TAP

    (JDs must be half-integer)

    :param jds: Half-integer JDs
    :return: None
    """
    with requests.Session() as session:
        # Create a session and do the login.
        # The cookie will end up in the cookie jar of the session.
        response = session.post(IPAC_TAP_URL, data=data, headers=headers)
        response.raise_for_status()
        auth = authsession.AuthSession()
        auth.credentials.set(securitymethods.ANONYMOUS, session)
        client = pyvo.dal.TAPService("https://irsa.ipac.caltech.edu/TAP", auth)

        for jd in jds:
            assert jd - int(jd) == 0.5, "JD must be a half-integer"

            obstable = client.search(
                f"""
            SELECT expid,obsjd,fid,field,exptime,rcid,maglimit,infobits,ipac_gid
            FROM ztf.ztf_current_meta_sci WHERE (obsjd BETWEEN {jd} AND {jd + 1})
            """
            ).to_table()
            names = (
                "expid",
                "exptime",
                "field",
                "fid",
                "rcid",
                "maglimit",
                "infobits",
                "ipac_gid",
            )
            renames = (
                "exposure_id",
                "exposure_time",
                "field_id",
                "filter_id",
                "qid",
                "maglim",
                "status",
                "programid",
            )
            obstable.rename_columns(names, renames)

            obs = obstable.to_pandas()
            obs["nalertpackets"] = np.nan

            output_path = coverage_tap_path(jd)

            obs.to_json(output_path)
            time.sleep(1)

            logger.warning(
                f"Coverage for JD {jd} written to {output_path} using TAP, "
                f"but this will only include programid=1"
            )


def write_coverage_skyvision(jds: list[float]):
    """
    Write the Skyvision coverage for a list of JDs to the cache

    :param jds: JDs
    :return: None
    """
    for jd in tqdm(jds):
        date = get_date(jd)

        assert jd - int(jd) == 0.5, "JD must be a half-integer"

        res = skyvision.get_log(f"{date[:4]}-{date[4:6]}-{date[6:8]}", verbose=False)

        path = coverage_skyvision_path(jd)

        if res is not None:
            # The skyvision log has a bug where some entries have a FieldID of "NONE"
            mask = (
                pd.notnull(res["FieldID"]) & res["FieldID"].astype(str).str.isnumeric()
            )
            res = res[mask]
            jds = [
                Time(f'{row["UT Date"]}T{row["UT Time"]}').jd
                for _, row in res.iterrows()
            ]
            res["obsjd"] = jds
            res["status"] = (res["Observation Status"] == "FAILED").astype(int)
            res["filter_id"] = res["Filter"].apply(
                lambda x: 1 if x == "FILTER_ZTF_G" else 2 if x == "FILTER_ZTF_R" else 3
            )
            res["maglim"] = 20.5
            res["field_id"] = res["FieldID"].astype(int)
            res["exposure_time"] = res["Exptime"]
            res = res[
                ["obsjd", "filter_id", "field_id", "exposure_time", "maglim", "status"]
            ]

            new_res = []
            for _, row in res.iterrows():
                new = row.to_dict()
                new_res += [dict(qid=int(i), **new) for i in range(64)]
            pd.DataFrame(new_res).to_json(path)

        else:
            with open(path, "w") as f:
                json.dump({}, f)


# The following functions are used to get the coverage from the cache


def get_coverage(jds: [int]) -> pd.DataFrame | None:
    """
    Get a dataframe of the coverage for a list of JDs

    Will use the cache if available, otherwise will query the depot, and lastly TAP

    :param jds: JDs
    :return: Coverage dataframe
    """
    # Clear any logs flagged as partial/incomplete

    cache_files = coverage_dir.glob("*.json")
    partial_logs = [x for x in cache_files if partial_flag in str(x)]

    if len(partial_logs) > 0:
        logger.debug(f"Removing the following partial logs: {partial_logs}")
        for partial_log in partial_logs:
            partial_log.unlink()

    # Only write missing logs

    missing_logs = []

    for jd in jds:
        depot_path = coverage_depot_path(jd)
        if not depot_path.exists():
            missing_logs.append(jd)
        else:
            df = pd.read_json(coverage_depot_path(jd))
            if len(df) == 0:
                missing_logs.append(jd)

    if len(missing_logs) > 0:
        logger.info(
            f"Some logs were missing from the cache. "
            f"Querying for the following JDs in depot: {missing_logs}"
        )
        write_coverage_depot(missing_logs)

    # Try skyvision for missing logs
    still_missing_logs = []
    for jd in missing_logs:
        skyvision_path = coverage_skyvision_path(jd)
        if not skyvision_path.exists():
            still_missing_logs.append(jd)
        else:
            df = pd.read_json(coverage_skyvision_path(jd))
            if len(df) == 0:
                still_missing_logs.append(jd)

    if len(still_missing_logs) > 0:
        logger.info(
            f"Some logs were still missing from the cache. "
            f"Querying for the following JDs from skyvision: {still_missing_logs}"
        )
        write_coverage_skyvision(still_missing_logs)

    # Try TAP for missing logs

    completely_missing_logs = []

    for jd in still_missing_logs:
        depot_path = coverage_depot_path(jd)
        if not depot_path.exists():
            completely_missing_logs.append(jd)
        else:
            df = pd.read_json(coverage_depot_path(jd))
            if len(df) == 0:
                completely_missing_logs.append(jd)

    if len(completely_missing_logs) > 0:
        logger.info(
            f"Some logs were still missing from the cache. "
            f"Querying for the following JDs from TAP: {completely_missing_logs}"
        )
        write_coverage_tap(completely_missing_logs)

    # Load logs from cache

    results = []

    for jd in tqdm(jds):
        res = pd.read_json(coverage_depot_path(jd))
        if len(res) > 0:
            results.append(res)
        else:
            res = pd.read_json(coverage_skyvision_path(jd))
            if len(res) > 0:
                results.append(res)
            else:
                res = pd.read_json(coverage_tap_path(jd))
                results.append(res)

    if results:
        result_df = pd.concat(results, ignore_index=True)
        return result_df
    else:
        return None


def get_obs_summary_depot(t_min: Time, t_max: Time) -> MNS | None:
    """
    Get observation summary from depot
    """

    jds = np.arange(int(t_min.jd) - 0.5, int(t_max.jd) + 1.5)

    res = get_coverage(jds)

    if len(res) == 0:
        return None

    res["date"] = Time(res["obsjd"].to_numpy(), format="jd").isot

    mns = MNS(df=res)

    mns.data.query(f"obsjd >= {t_min.jd} and obsjd <= {t_max.jd}", inplace=True)

    mns.data.reset_index(inplace=True)
    mns.data.drop(columns=["index"], inplace=True)

    return mns


def get_obs_summary_skyvision(t_min, t_max):
    """
    Get observation summary from Skyvision
    """
    t_min_date = t_min.to_value("iso", subfmt="date")
    t_max_date = t_max.to_value("iso", subfmt="date")

    # ztfquery saves nightly observations in a cache, and does not redownload them.
    # If the nightly log was not complete, it will never be updated.
    # Here we simply clear the cache and cleanly re-download everything.

    logger.debug(
        f"Skyvision: Obtaining nightly observation logs from {t_min_date} to {t_max_date}"
    )

    skyvision_log = os.path.join(LOCALSOURCE, "skyvision")

    for filename in os.listdir(skyvision_log):
        if ".csv" in filename:
            path = os.path.join(skyvision_log, filename)
            os.remove(path)

    mns = skyvision.CompletedLog.from_daterange(
        t_min_date, end=t_max_date, verbose=False
    )

    mns.data["obsjd"] = Time(list(mns.data.datetime.values), format="isot").jd

    mns.data.query(f"obsjd >= {t_min.jd} and obsjd <= {t_max.jd}", inplace=True)

    mns.data.reset_index(inplace=True)
    mns.data.drop(columns=["index"], inplace=True)

    logger.debug(f"Found {len(mns.data)} observations in total.")

    return mns


def get_obs_summary(t_min, t_max=None, max_days: float = None) -> MNS | None:
    """
    Get observation summary from IPAC depot
    """
    now = Time.now()

    if t_max and max_days:
        raise ValueError("Choose either t_max or max_days, not both")

    if t_max is None:
        if max_days is None:
            t_max = now
        else:
            t_max = t_min + (max_days * u.day)

    if t_max > now:
        t_max = now

    logger.info("Getting observation logs from IPAC depot.")
    mns = get_obs_summary_depot(t_min=t_min, t_max=t_max)

    if mns is not None:
        logger.info(f"Found {len(set(mns.data))} observations in total.")
    else:
        logger.warning("Found no observations on IPAC depot or TAP.")

    return mns



#!/usr/bin/env python3
# Author: Simeon Reusch (simeon.reusch@desy.de)
# License: BSD-3-Clause

import argparse
import logging
import os
import re

# from ampel.ztf.archive.ArchiveDB import ArchiveDB
import ampel.contrib.hu.t3.TNSTalker as TNSTalker

# from ampel.ztf.pipeline.common.ZTFUtils import ZTFUtils
from ampel.contrib.hu.t3.ampel_tns import sendTNSreports
from ampel.ztf import legacy_utils

# from ampel.ztf.utils.ZIAlertUtils import ZIAlertUtils
from ampel.ztf.dev.ZTFAlert import ZTFAlert

from nuztf.credentials import load_credentials


def TNSSubmit(ztf_ids: list, reporter: str = None, sandbox: bool = True, logger=None):
    """ """
    if logger is None:
        logger = logging.getLogger(__name__)

    _, tns_api_token = load_credentials("tns_api_token")

    # Run config for TNSTalker
    ztf_default_values = {
        "flux_units": "1",
        "instrument_value": "196",
        "Observer": "Robot",
    }

    RUN_CONFIG = {
        "submit_tns": True,
        "sandbox": True,
        "resubmit_tns_nonztf": False,
        "resubmit_tns_ztf": False,
        "submit_unless_journal": True,
        "lc_filters": [],
        "ztf_tns_at": ztf_default_values,
        "max_age": 5,
        "tns_api_key": tns_api_token,
    }

    # Connect to TNSTalker instance
    talker = TNSTalker.TNSTalker(logger=logger, run_config=RUN_CONFIG)
    talker.run_config = talker.RunConfig(**RUN_CONFIG)


# # Connect to TNSTalker instance
# TNS = TNSTalker.TNSTalker(logger=logger, run_config=RUN_CONFIG)
# TNS.run_config = TNS.RunConfig(**RUN_CONFIG)
# PORT = 5432

# # Connect to AMPEL instance
# AMPEL_CLIENT = ArchiveDB(
#     "postgresql://{0}:{1}@localhost:{2}/ztfarchive".format(username, password, PORT)
# )


# def is_ztf_name(name):
#     """ """
#     return re.match("^ZTF[1-2]\d[a-z]{7}$", name)


# def use_if_ztf(file_or_name):
#     """ """
#     if is_ztf_name(file_or_name):
#         object_list = [file_or_name]
#     else:
#         object_list = []
#         try:
#             file = open(f"{file_or_name}", "r")
#             lines = file.read().splitlines()
#             for line in lines:
#                 if is_ztf_name(line):
#                     object_list.append(line)
#         except FileNotFoundError as error:
#             print(
#                 "\nYou have to provide either a ZTF name or a file containing ZTF names (1 per line).\n"
#             )
#             raise error
#         assert (
#             object_list[0][:3] == "ZTF" and len(object_list[0]) == 12
#         ), "You have to provide either a ZTF name or a file containing ZTF names (1 per line)"

#     return object_list


# def TNS_submit(ztf_names, reporter=None, sandbox=True):
#     """
#     Note: IF YOU WANT TO SUBMIT, SET sandbox=False
#     """
#     atreports = {}

#     for ztf_name in ztf_names:
#         ztf_object = AMPEL_CLIENT.get_alerts_for_object(ztf_name, with_history=True)
#         query_res = [i for i in ztf_object]
#         last_alert = query_res[-1]
#         tv = ZIAlertUtils.to_transientview(
#             file_path=None, content=last_alert, science_records=None
#         )
#         ztf_id = tv.tran_id
#         ampel_id = ZTFUtils.to_ampel_id(ztf_id)
#         lc_id = tv.lightcurves[0].id
#         object.__setattr__(tv, "tran_id", ampel_id)
#         object.__setattr__(tv, "latest_state", lc_id)
#         atreport = TNS.create_atreport(tv)
#         if reporter is None:
#             atreport["reporter"] = "The Zwicky Transient Facility (ZTF) Collaboration"
#         else:
#             atreport[
#                 "reporter"
#             ] = f"{reporter} for the Zwicky Transient Facility (ZTF) Collaboration"
#         atreport["exptime"] = "300"
#         atreports.update({ztf_name: atreport})

#     # Reformat the report (code by AMPEL!)
#     atreportlist = []
#     atreport = {}
#     k = 0
#     for tranid in atreports.keys():
#         if len(atreport) > 90:
#             self.logger.info("adding another report to TNS submit")
#             atreportlist.append({"at_report": atreport})
#             atreport = {}
#             k = 0
#         atreport[int(k)] = atreports[tranid]
#         k += 1
#     atreportlist.append({"at_report": atreport})
#     tnsreplies = sendTNSreports(
#         atreportlist, TNS.run_config.tns_api_key, logger, sandbox=sandbox
#     )


# if __name__ == "__main__":
#     parser = argparse.ArgumentParser(description="Submit ZTF object(s) to TNS")
#     parser.add_argument(
#         "name",
#         type=str,
#         help='Provide a ZTF name (e.g. "ZTF19aaelulu") or a .txt-file containing a list of ZTF names',
#     )
#     parser.add_argument(
#         "--commit",
#         "-commit",
#         action="store_false",
#         help="Run an actual commit (instead of just using the TNS sandbox",
#     )
#     parser.add_argument(
#         "--reporter",
#         "-reporter",
#         type=str,
#         default=None,
#         help="Provide an author for the submission",
#     )

#     commandline_args = parser.parse_args()
#     name = commandline_args.name
#     sandbox = commandline_args.commit
#     reporter = commandline_args.reporter
#     object_list = use_if_ztf(name)

#     if sandbox:
#         logger.info(
#             f"Submitting {object_list} to the TNS - SANDBOX MODE ONLY (to change, use -commit)"
#         )
#     else:
#         logger.info(f"Submitting {object_list} to the TNS")

#     TNS_submit(object_list, reporter, sandbox=sandbox)



#!/usr/bin/env python
# coding: utf-8

import json
import logging
import os
import time
from pathlib import Path

import healpy as hp
import lxml.etree
import numpy as np
import requests
import wget
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.time import Time
from astropy_healpix import HEALPix
from ligo.gracedb.exceptions import HTTPError
from ligo.gracedb.rest import GraceDb
from ligo.skymap.io import read_sky_map
from ligo.skymap.moc import rasterize
from lxml import html
from numpy.lib.recfunctions import append_fields

from nuztf.paths import RESULTS_DIR, SKYMAP_DIR


class EventNotFound(Exception):
    """Base class for non-existing event"""


class RetractionError(Exception):
    """Base class for retracted event"""


class Skymap:
    def __init__(
        self,
        event: str = None,
        rev: int = None,
        prob_threshold: float = 0.9,
        output_nside: int | None = None,
    ):
        self.logger = logging.getLogger(__name__)

        self.prob_threshold = prob_threshold
        self.event = event

        if ".fit" in event:
            basename = Path(event).stem

            self.event_name = basename
            self.skymap_path = SKYMAP_DIR.joinpath(basename)
            self.rev = None

            if event[:8] == "https://":
                if not self.skymap_path.exists():
                    self.logger.info(f"Downloading from: {event}")
                    wget.download(event, str(self.skymap_path))
                else:
                    self.logger.info(f"Found saved skymap at: {self.skymap_path}")
        #
        # elif "grb" in event.lower():
        #     self.skymap_path, self.event_name = self.get_grb_skymap(event=event)

        else:
            self.skymap_path, self.event_name, self.rev = self.get_gw_skymap(
                event=event, rev=rev
            )

        (
            self.data,
            self.t_obs,
            self.hpm,
            self.key,
            self.dist,
            self.dist_unc,
        ) = self.read_map(output_nside=output_nside)

        t_min = Time(self.t_obs, format="isot", scale="utc")

        self.logger.info(f"Event time: {t_min}")
        self.logger.info("Reading map")

        self.pixel_threshold = self.find_pixel_threshold(self.data[self.key])

    #
    # def get_grb_skymap(self, event: str):
    #     """ """
    #     if event is None:
    #         raise ValueError(
    #             "event_name must be provided for GRBs. They must have the form 'GRB210729A"
    #         )
    #
    #     event_year_short = event[3:5]
    #     event_year = "20" + event_year_short
    #     event_month = event[5:7]
    #     event_day = event[7:9]
    #     event_letter = event[9]
    #     event_number = ord(event_letter) - 65
    #
    #     # get possible skymap URLs
    #
    #     url = f"https://heasarc.gsfc.nasa.gov/FTP/fermi/data/gbm/triggers/{event_year}"
    #
    #     page_overview = requests.get(url)
    #     webpage_overview = html.fromstring(page_overview.content)
    #
    #     links_overview = webpage_overview.xpath("//a/@href")
    #
    #     links_for_date = []
    #
    #     for link in links_overview:
    #         if link[2:8] == f"{event_year_short}{event_month}{event_day}":
    #             links_for_date.append(url + "/" + link + "current/")
    #
    #     if len(links_for_date) > 1:
    #         self.logger.info(
    #             f"Found multiple events. "
    #             f"Will choose the one corresponding the GRB letter {event_letter}"
    #         )
    #
    #     event_url = links_for_date[event_number]
    #
    #     page_event = requests.get(event_url)
    #     webpage_event = html.fromstring(page_event.content)
    #     links_event = webpage_event.xpath("//a/@href")
    #
    #     for link in links_event:
    #         if link[0:11] == "glg_healpix":
    #             final_link = event_url + link
    #             break
    #
    #     self.skymap_path = os.path.join(self.base_skymap_dir, link)
    #
    #     if os.path.isfile(self.skymap_path):
    #         self.logger.info(
    #             f"Continuing with saved skymap. Located at {self.skymap_path}"
    #         )
    #     else:
    #         self.logger.info(f"Downloading skymap and saving to {self.skymap_path}")
    #         wget.download(final_link, self.skymap_path)
    #
    #     return sk, event

    def get_gw_skymap(self, event: str, rev: int | None = None):
        """
        Function to download a GW skymap from GraceDB

        :param event: Event name
        :param rev: revision number
        :return:
        """

        ligo_client = GraceDb()

        self.logger.info("Obtaining skymap from GraceDB")

        if event is None:
            superevent_iterator = ligo_client.superevents("category: Production")
            superevent_ids = [
                superevent["superevent_id"] for superevent in superevent_iterator
            ]
            event = superevent_ids[0]

        try:
            res = ligo_client.voevents(event)
            if res.status_code == 200:
                voevents = res.json()["voevents"]
            else:
                raise EventNotFound(
                    f"The specified LIGO event, {event}, was not found on GraceDB. "
                    f"Please check that you entered the correct event name."
                )
        except HTTPError:
            raise EventNotFound(
                f"The specified LIGO event, {event}, was not found on GraceDB. "
                f"Please check that you entered the correct event name."
            )

        if rev is None:
            rev = len(voevents)

        elif rev > len(voevents):
            raise Exception("Revision {0} not found".format(rev))

        latest_voevent = voevents[rev - 1]
        self.logger.info(f"Found voevent {latest_voevent['filename']}")

        if "Retraction" in latest_voevent["filename"]:
            raise RetractionError(
                f"The specified LIGO event, "
                f"{latest_voevent['filename']}, was retracted."
            )

        response = requests.get(latest_voevent["links"]["file"])

        root = lxml.etree.fromstring(response.content)
        params = {
            elem.attrib["name"]: elem.attrib["value"]
            for elem in root.iterfind(".//Param")
        }

        latest_skymap = params["skymap_fits"]

        self.logger.info(f"Latest skymap URL: {latest_skymap}")

        savepath = SKYMAP_DIR.joinpath(
            f"{event}_rev{latest_voevent['N']}_{os.path.basename(latest_skymap)}"
        )

        self.logger.info(f"Saving to: {savepath}")
        response = requests.get(latest_skymap)

        with open(savepath, "wb") as f:
            f.write(response.content)

        event_name = f"{event}/rev{latest_voevent['N']}"

        return savepath, event_name, rev

    def read_map(self, output_nside: int | None = None):
        """
        Read the skymap

        :param output_nside: The nside of the output skymap.
            If None, the nside of the input skymap will be used.

        """

        self.logger.info(f"Reading file: {self.skymap_path}")

        with fits.open(self.skymap_path) as hdul:
            data = None
            h = hdul[0].header

            dist = None
            dist_unc = None
            t_obs = None
            ordering = None

            for x in hdul:
                if data is None:
                    if x.data is not None:
                        data = np.array(x.data)

                if "DISTMEAN" in x.header:
                    dist = x.header["DISTMEAN"]

                if "DISTSTD" in x.header:
                    dist_unc = x.header["DISTSTD"]

                if "DATE-OBS" in x.header:
                    t_obs = x.header["DATE-OBS"]

                elif "EVENTMJD" in x.header:
                    t_obs_mjd = x.header["EVENTMJD"]
                    t_obs = Time(t_obs_mjd, format="mjd").isot

                if "ORDERING" in x.header:
                    ordering = x.header["ORDERING"]

        if "PROB" in data.dtype.names:
            pass
        elif (replacekey := "PROBABILITY") in data.dtype.names:
            prob = np.array(data[replacekey]).flatten()
            data = append_fields(data, "PROB", prob)
        elif (replacekey := "PROBDENSITY") in data.dtype.names:
            prob = np.array(data[replacekey])
            data = append_fields(data, "PROB", prob)
        elif (replacekey := "T") in data.dtype.names:  # weird IceCube format
            prob = np.array(data[replacekey]).flatten()
            data = append_fields(data, "PROB", prob)
        else:
            raise Exception(
                f"No recognised probability key in map. This is probably a weird one, right? "
                f"Found the following keys: {data.dtype.names}"
            )

        if ordering == "NUNIQ":
            self.logger.info("Rasterising skymap to convert to nested format")
            # We need to use the ligo.skymap.io map parser to make rasterize work
            skymap_uniq = read_sky_map(self.skymap_path, moc=True)
            if (replacekey := "PROBDENSITY") in skymap_uniq.colnames:
                skymap_uniq[replacekey].name = "PROB"

            data = rasterize(skymap_uniq, order=7)

        if not isinstance(data["PROB"][0], float):
            self.logger.info("Flattening skymap")
            prob = np.array(data["PROB"]).flatten()
            data = np.array(prob, dtype=np.dtype([("PROB", float)]))

        if "NSIDE" not in h.keys():
            h["NSIDE"] = hp.npix2nside(len(data["PROB"]))

        data["PROB"] /= np.sum(data["PROB"])

        self.logger.info(f"Summed probability is {100. * np.sum(data['PROB']):.1f}%")

        if ordering == "RING":
            data["PROB"] = hp.pixelfunc.reorder(data["PROB"], inp="RING", out="NESTED")

        if ordering is not None:
            h["ORDERING"] = "NESTED"

        else:
            raise Exception(
                f"Error parsing fits file, no ordering found. "
                f"Please enter the ordering (NESTED/RING/NUNIQ)"
            )

        # Optionally interpolate to a different nside
        if output_nside is not None:
            if output_nside != hp.npix2nside(len(data["PROB"])):
                self.logger.info(f"Regridding to nside {output_nside}")

                new_prob = hp.ud_grade(
                    data["PROB"],
                    nside_out=output_nside,
                    order_in=h["ORDERING"],
                    order_out="NESTED",
                    power=-2,
                )

                new_data = np.array(new_prob, dtype=np.dtype([("PROB", float)]))

                if "DISTMEAN" in h.keys():
                    dist = hp.ud_grade(
                        data["DISTMEAN"],
                        nside_out=output_nside,
                        order_in=h["ORDERING"],
                        order_out="NESTED",
                        power=0,
                    )

                    new_data["DISTMEAN"] = dist

                if "DISTSTD" in h.keys():
                    dist_unc = hp.ud_grade(
                        data["DISTSTD"],
                        nside_out=output_nside,
                        order_in=h["ORDERING"],
                        order_out="NESTED",
                        power=0,
                    )
                    new_data["DISTSTD"] = dist_unc

                data = new_data

        else:
            output_nside = h["NSIDE"]

        hpm = HEALPix(nside=output_nside, order="NESTED", frame="icrs")

        return data, t_obs, hpm, "PROB", dist, dist_unc

    def find_pixel_threshold(self, data):
        """
        Find the pixel threshold that corresponds to the probability threshold

        :param data: Healpix map data
        :return: threshold value
        """

        ranked_pixels = np.sort(data)[::-1]
        int_sum = 0.0
        pixel_threshold = 0.0

        for i, prob in enumerate(ranked_pixels):
            int_sum += prob
            if int_sum > self.prob_threshold:
                self.logger.info(
                    f"Threshold found! \n To reach {int_sum * 100.0:.2f}% of "
                    f"probability, pixels with probability greater "
                    f"than {prob:.3g} are included."
                )
                pixel_threshold = prob
                break

        return pixel_threshold

    def interpolate_map(self, ra_deg: float, dec_deg: float) -> float:
        """
        Find the probability at a given sky position by interpolating the skymap

        :param ra_deg: Right ascension in degrees
        :param dec_deg: declination in degrees
        :return: Value of the skymap at the given position
        """
        interpol_map = self.hpm.interpolate_bilinear_skycoord(
            SkyCoord(ra_deg * u.deg, dec_deg * u.deg), self.data[self.key]
        )
        return interpol_map

    def in_contour(self, ra_deg: float, dec_deg: float) -> bool:
        """
        Find if a given sky position is within the contour of the skymap

        :param ra_deg: Right ascension in degrees
        :param dec_deg: declination in degrees
        :return: bool
        """
        return self.interpolate_map(ra_deg, dec_deg) > self.pixel_threshold



#!/usr/bin/env python
# coding: utf-8

import json
import logging
import warnings

from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.utils.exceptions import AstropyWarning
from astroquery.exceptions import RemoteServiceError
from astroquery.ipac.irsa import Irsa
from astroquery.ipac.ned import Ned

from nuztf.ampel_api import ampel_api_catalog, ampel_api_name
from nuztf.paths import CROSSMATCH_CACHE


def query_ned_for_z(
    ra_deg: float, dec_deg: float, searchradius_arcsec: float = 20, logger=None
):
    """Function to obtain redshifts from NED (via the AMPEL API)"""

    z = None
    dist_arcsec = None

    query = ampel_api_catalog(
        catalog="NEDz_extcats",
        catalog_type="extcats",
        ra_deg=ra_deg,
        dec_deg=dec_deg,
        search_radius_arcsec=searchradius_arcsec,
        search_type="nearest",
        logger=logger,
    )

    if query:
        z = query["body"]["z"]
        dist_arcsec = query["dist_arcsec"]

    return z, dist_arcsec


def query_ned_astroquery(
    ra_deg: float, dec_deg: float, searchradius_arcsec: float = 0.5
):
    """
    Function to obtain NED crossmatches via astroquery
    """
    c = SkyCoord(ra_deg, dec_deg, unit=u.deg, frame="icrs")

    r = searchradius_arcsec * u.arcsecond

    try:
        return Ned.query_region(c, radius=r)
    except RemoteServiceError:
        return None


def query_wise_astroquery(
    ra_deg: float, dec_deg: float, searchradius_arcsec: float = 3.0
):
    """
    Function to obtain WISE crossmatches via astroquery

    :param ra_deg: Right ascension (deg)
    :param dec_deg: Declination (deg)
    :param searchradius_arcsec: Search radius (arcsec)
    :return: result of query
    """
    c = SkyCoord(ra_deg, dec_deg, unit=u.deg, frame="icrs")

    r = searchradius_arcsec * u.arcsecond

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        allwise = Irsa.query_region(c, catalog="allwise_p3as_psd", radius=r)
    return allwise


def ampel_api_tns(
    ra_deg: float, dec_deg: float, searchradius_arcsec: float = 3, logger=None
):
    """Function to query TNS via the AMPEL API"""

    full_name = None
    discovery_date = None
    source_group = None

    res = ampel_api_catalog(
        catalog="TNS",
        catalog_type="extcats",
        ra_deg=ra_deg,
        dec_deg=dec_deg,
        search_radius_arcsec=searchradius_arcsec,
        search_type="nearest",
        logger=logger,
    )

    if res:
        response_body = res["body"]
        name = response_body["objname"]
        prefix = response_body["name_prefix"]
        full_name = prefix + name
        discovery_date = response_body["discoverydate"]
        if "source_group" in response_body.keys():
            source_group = response_body["source_group"]["group_name"]

    return full_name, discovery_date, source_group


def get_cross_match_info(raw: dict, logger=None):
    """ """

    cache_file = CROSSMATCH_CACHE.joinpath(f"{raw['objectId']}.json")

    if cache_file.exists():
        with open(cache_file) as f:
            res = json.load(f)
            label = res["data"]
            return label

    alert = raw["candidate"]

    label = ""

    if logger is None:
        logger = logging.getLogger()

    # Check if known variable star (https://arxiv.org/pdf/1405.4290.pdf)

    res = ampel_api_catalog(
        catalog="CRTS_DR1",
        catalog_type="extcats",
        ra_deg=alert["ra"],
        dec_deg=alert["dec"],
        search_radius_arcsec=5.0,
        logger=logger,
    )
    if res is not None:
        if logger:
            logger.info(res)
        label = (
            f"[CRTS variable star: "
            f"{res[0]['body']['name']} ({res[0]['dist_arcsec']:.2f} arsec)]"
        )

    # Check if known QSO/AGN

    res = ampel_api_catalog(
        catalog="milliquas",
        catalog_type="extcats",
        ra_deg=alert["ra"],
        dec_deg=alert["dec"],
        search_radius_arcsec=1.5,
        logger=logger,
    )
    if res is not None:
        if len(res) == 1:
            if "q" in res[0]["body"]["broad_type"]:
                label = (
                    f"[MILLIQUAS: {res[0]['body']['name']} - "
                    f"Likely QSO (prob = {res[0]['body']['qso_prob']}%) "
                    f"({res[0]['dist_arcsec']:.2f} arsec)]"
                )
            else:
                label = (
                    f"[MILLIQUAS: {res[0]['body']['name']} - "
                    f"'{res[0]['body']['broad_type']}'-type source "
                    f"({res[0]['dist_arcsec']:.2f} arsec)]"
                )
        else:
            label = "[MULTIPLE MILLIQUAS MATCHES]"

    # Check if measured parallax in Gaia (i.e galactic)

    if label == "":
        res = ampel_api_catalog(
            catalog="GAIADR2",
            catalog_type="catsHTM",
            ra_deg=alert["ra"],
            dec_deg=alert["dec"],
            search_radius_arcsec=5.0,
            logger=logger,
        )
        if res is not None:
            if res[0]["body"]["Plx"] is not None:
                plx_sig = res[0]["body"]["Plx"] / res[0]["body"]["ErrPlx"]
                if plx_sig > 3.0:
                    label = (
                        f"[GAIADR2: {plx_sig:.1f}-sigma parallax "
                        f"({res[0]['dist_arcsec']:.2f} arsec)]"
                    )

    # Check if classified as probable star in SDSS

    if label == "":
        res = ampel_api_catalog(
            catalog="SDSSDR10",
            catalog_type="catsHTM",
            ra_deg=alert["ra"],
            dec_deg=alert["dec"],
            search_radius_arcsec=1.5,
            logger=logger,
        )
        if res is not None:
            if len(res) == 1:
                if float(res[0]["body"]["type"]) == 6.0:
                    label = (
                        f"[SDSS Morphology: 'Star'-type source "
                        f"({res[0]['dist_arcsec']:.2f} arsec)]"
                    )
            else:
                label = "[MULTIPLE SDSS MATCHES]"

    # WISE colour cuts (https://iopscience.iop.org/article/10.3847/1538-4365/)

    if label == "":
        res = query_wise_astroquery(
            ra_deg=alert["ra"],
            dec_deg=alert["dec"],
            searchradius_arcsec=3.0,
        )
        if res is not None:
            if len(res) > 0:
                w1mw2 = res["w1mpro"][0] - res["w2mpro"][0]

                if w1mw2 > 0.8:
                    label = (
                        f"[Probable WISE-selected quasar:W1-W2={w1mw2:.2f}>0.8  "
                        f"({res[0]['dist']:.2f} arsec)]"
                    )
                elif w1mw2 > 0.5:
                    label = (
                        f"[Possible WISE-selected quasar:W1-W2={w1mw2:.2f}>0.5  "
                        f"({res[0]['dist']:.2f} arsec)]"
                    )
                else:
                    label = (
                        f"WISE DETECTION: W1-W2={w1mw2:.2f} "
                        f"({res[0]['dist']:.2f} arsec)"
                    )

                if len(res) > 1:
                    label += "[MULTIPLE WISE MATCHES]"

    # Just check NED

    if label == "":
        res = query_ned_astroquery(
            ra_deg=alert["ra"],
            dec_deg=alert["dec"],
            searchradius_arcsec=3.0,
        )

        if res is not None:
            if len(res) == 1:
                label = (
                    f"{res['Object Name'][0]} ['{res['Type'][0]}'-type source "
                    f"({res['Separation'][0]:.2f} arsec)]"
                )
            elif len(res) > 1:
                label += "[MULTIPLE NED MATCHES]"
                logger.debug(f"{res}")

    # Extra check to TNS, append to other info

    full_name, _, _ = ampel_api_tns(
        ra_deg=alert["ra"],
        dec_deg=alert["dec"],
    )

    if full_name is not None:
        label += f" [TNS NAME={full_name}]"

    with open(cache_file, "w") as f:
        json.dump({"data": label}, f)

    return label


def check_cross_match_info_by_name(name: str, logger=None):
    """
    Utility function to check cross-match info for a given name

    :param name: ZTF name
    :param logger:
    :return:
    """
    return get_cross_match_info(
        raw=ampel_api_name(name, with_history=False, logger=logger)[0], logger=logger
    )



#!/usr/bin/env python3
# coding: utf-8

import json
import logging
import math
import re
from collections import OrderedDict, defaultdict
from json import JSONDecodeError

import numpy as np
import requests
from astropy.cosmology import FlatLambdaCDM
from requests.auth import HTTPBasicAuth

from nuztf.credentials import load_credentials

# same cosmology everywhere
cosmo = FlatLambdaCDM(H0=70, Om0=0.3)


def is_icecube_name(name) -> bool:
    """
    Checks if a string adheres to the IceCube naming scheme
    (e.g. IC201021B)
    """
    if re.match(
        r"^IC((\d{2}((0[13578]|1[02])(0[1-9]|[12]\d|3[01])|(0[13456789]|1[012])(0[1-9]|[12]\d|30)|02(0[1-9]|1\d|2[0-8])))|([02468][048]|[13579][26])0229)[a-zA-Z]$",
        name,
    ):
        match = True
    else:
        match = False
    return match


def is_ligo_name(name) -> bool:
    """
    Checks if a string adheres to the LVT naming scheme
    (e.g. S190814bv)
    """
    if re.match(
        r"^(S|GW|MS)((\d{2}((0[13578]|1[02])(0[1-9]|[12]\d|3[01])|(0[13456789]|1[012])(0[1-9]|[12]\d|30)|02(0[1-9]|1\d|2[0-8])))|([02468][048]|[13579][26])0229)[a-z]{0,4}$",
        name,
    ):
        match = True
    else:
        match = False
    return match


def is_ztf_name(name: str) -> bool:
    """
    Checks if a string adheres to the ZTF naming scheme
    """
    if re.match(r"^ZTF[1-2]\d[a-z]{7}$", name):
        match = True
    else:
        match = False
    return match


def is_tns_name(name: str) -> bool:
    """
    Checks if a string adheres to the TNS naming scheme
    """
    if re.match(r"^AT|SN(19|20)\d\d[a-z]{3,4}$", name):
        matches = True
    else:
        matches = False
    return matches


def reformat_downloaded_results(photopoints: list, ztf_id: str) -> dict:
    """
    Massage the TransientView photopoint output so it matches what the archive DB returns
    """

    # Find the index of the latest detection
    latest_jd_index = np.argmax(np.asarray([pp["body"]["jd"] for pp in photopoints]))

    resdict = {
        "candid": photopoints[latest_jd_index]["id"],
        "objectId": ztf_id,
        "schemavsn": "3.3",
        "publisher": "Ampel",
        "candidate": photopoints[latest_jd_index]["body"],
    }

    prv_candidates = []

    for i, pp in enumerate(photopoints):
        if i != latest_jd_index:  # dict is readonly, we can't pop
            prv_candidates.append(pp["body"])

    resdict["prv_candidates"] = prv_candidates

    return resdict


def deres(nside, ipix, min_nside=1):
    """
    Originally from Ampel-ZTF-archive/ampel/ztf/archive/server (by JvS)

    Decompose a set of (nested) HEALpix indices into sets of complete superpixels at lower resolutions.
    :param nside: nside of given indices
    :param ipix: pixel indices
    :min_nside: minimum nside of complete pixels
    """
    remaining_pixels = set(ipix)
    decomposed = defaultdict(list)
    for log2_nside in range(int(math.log2(min_nside)), int(math.log2(nside)) + 1):
        super_nside = 2**log2_nside
        # number of base_nside pixels per nside superpixel
        scale = (nside // super_nside) ** 2
        # sort remaining base_nside pixels by superpixel
        by_superpixel = defaultdict(list)
        for pix in remaining_pixels:
            by_superpixel[pix // scale].append(pix)
        # represent sets of pixels that fill a superpixel
        # as a single superpixel, and remove from the working set
        for superpix, members in by_superpixel.items():
            if len(members) == scale:
                decomposed[super_nside].append(superpix)
                remaining_pixels.difference_update(members)

    decomposed_dict = dict(decomposed)

    healpix_regions = [
        {"nside": nside, "pixels": pixels} for nside, pixels in decomposed_dict.items()
    ]

    return healpix_regions


def query_tns_by_name(name, logger=None):
    """
    Query the TNS API for a given name
    """
    if not is_tns_name(name):
        raise ValueError("String is not a TNS name")

    tns_api_token = load_credentials("tns_api_token", token_based=True)

    name = name[2:]

    tns_bot_id = "115364"
    tns_bot_name = "ZTF_DESY"

    if logger is None:
        logger = logging.getLogger(__name__)

    # logger = logging.getLogger(__name__)
    logger.info(f"Obtaining TNS information for {name}.")

    queryurl_tns = "https://www.wis-tns.org/api/get/object"

    tns_marker = (
        'tns_marker{"tns_id": "'
        + str(tns_bot_id)
        + '", "type": "bot", "name": "'
        + tns_bot_name
        + '"}'
    )
    headers = {"User-Agent": tns_marker}

    get_obj = [("objname", name), ("objid", ""), ("photometry", "1"), ("spectra", "0")]
    json_file = OrderedDict(get_obj)

    get_data = {"api_key": tns_api_token, "data": json.dumps(json_file)}

    response = requests.post(queryurl_tns, headers=headers, data=get_data)

    try:
        res = response.json()
        if "name" in res["data"]["reply"].keys():
            if "110" in res["data"]["reply"]["name"].keys():
                return None
        else:
            return res

    except JSONDecodeError:
        if response.headers:
            logger.debug(response.headers)
        raise requests.exceptions.RequestException



#!/usr/bin/env python3
# coding: utf-8

import logging
import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from astropy import constants as const
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table
from astropy.time import Time
from astroquery.exceptions import RemoteServiceError
from astroquery.ipac.ned import Ned
from ztfquery.io import LOCALSOURCE
from ztfquery.lightcurve import LCQuery

from nuztf.ampel_api import ampel_api_name
from nuztf.parse_nu_gcn import find_gcn_no, parse_gcn_circular
from nuztf.style import base_height, base_width, big_fontsize, dpi, plot_dir
from nuztf.utils import cosmo, is_tns_name, is_ztf_name, query_tns_by_name

logger = logging.getLogger(__name__)


def format_date(t, atel=True):
    t.format = "fits"

    if atel:
        frac_days = f"{t.mjd - int(t.mjd):.2f}"[1:]
        t.out_subfmt = "date"

        dt = "".join([t.value, frac_days])
    else:
        dt = t.value

    return dt


def load_irsa(ra_deg: float, dec_deg: float, radius_arcsec: float = 0.5, **kwargs):
    """
    Get lightcuve from IPAC
    """

    logger.debug("Querying IPAC")
    df = LCQuery.from_position(ra_deg, dec_deg, radius_arcsec, **kwargs).data

    logger.debug(f"Found {len(df)} datapoints")

    if len(df) == 0:
        logger.info("No data found.")
        return df

    else:
        mask = df.catflags > 0

        flags = list(set(df.catflags))

        logger.info(
            f"Found {len(df)} datapoints, masking {np.sum(mask)} datapoints with bad flags."
        )

        for flag in sorted(flags):
            logger.debug(f"{np.sum(df.catflags == flag)} datapoints with flag {flag}")

        df = df.drop(df[mask].index)
        return df


def plot_irsa_lightcurve(
    source_name: str,
    nu_name: list = None,
    source_coords: list = None,
    source_redshift: float = None,
    plot_mag: bool = False,
    atel: bool = True,
    plot_folder: str = plot_dir,
    extra_folder: str = None,
    check_obs: bool = True,
    check_obs_lookback_weeks: float = 4,
    from_cache: bool = False,
    cache_dir: str = os.path.join(LOCALSOURCE, "cache/"),
    expanded_labels: bool = True,
    ylim: tuple = None,
    radius_arcsec: float = 0.5,
    query_irsa_for_logs: bool = True,
) -> None:
    plot_title = source_name

    # If there are no coordinates, try name resolve to get coordinates!

    if source_coords is None:
        logger.info(f"Trying to resolve {source_name} and query for coordinates.")
        logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.INFO)

        # Try ampel to find ZTF coordinates

        if is_ztf_name(name=source_name):
            logger.info("Source name is a ZTF name.")
            res = ampel_api_name(source_name, with_history=False)[0]
            source_coords = [res["candidate"]["ra"], res["candidate"]["dec"]]
            logger.info(f"Found ZTF coordinates for source {source_name}")

        # Try TNS

        elif is_tns_name(name=source_name):
            logger.info("Source name is a TNS name.")
            result_dict = query_tns_by_name(name=source_name)

            if not result_dict:
                logger.warning(f"{source_name} is not in TNS.")

            if result_dict:
                logger.info(f"Found {source_name} on TNS.")
                res = result_dict["data"]["reply"]
                ra = res["radeg"]
                dec = res["decdeg"]
                source_coords = [ra, dec]
                if np.logical_and("redshift" in res.keys(), source_redshift is None):
                    source_redshift = res["redshift"]

        # Otherwise try NED

        else:
            logger.info(
                "Source name is neither as ZTF, nor a TNS name. Querying NED instead."
            )
            result_table = Ned.query_object(source_name)

            if len(result_table) == 0:
                logger.warning(
                    f"Failed to resolve name {source_name} in NED. Trying to be clever instead."
                )

                querystring = "".join(
                    [
                        x
                        for x in source_name
                        if x in [str(i) for i in range(10)] + ["+", "-"]
                    ]
                )

                result_table = Ned.query_object(
                    "".join(
                        [
                            x
                            for x in source_name
                            if x in [str(i) for i in range(10)] + ["+", "-"]
                        ]
                    )
                )

            if len(result_table) == 1:
                source_coords = [result_table["RA"][0], result_table["DEC"][0]]

                if "ZTF" in plot_title:
                    plot_title += f' ({result_table["Object Name"][0]})'

                if np.logical_and(
                    str(result_table["Redshift"][0]) != "--", source_redshift is None
                ):
                    source_redshift = result_table["Redshift"].data[0]

                logger.info(
                    f"Using AStroquery NED query result for name {source_name} ({source_coords})"
                )

            if source_coords is None:
                sc = SkyCoord.from_name(source_name)
                logger.info(
                    f"Using Astroquery CDS query result for name {source_name} (RA={sc.ra}, Dec={sc.dec})"
                )
                source_coords = (sc.ra.value, sc.dec.value)

    # Try to find a catalogue source nearby using coordinates

    if ("ZTF" in source_name) and source_coords:
        c = SkyCoord(source_coords[0], source_coords[1], unit=u.deg, frame="icrs")

        r = 0.5 * u.arcsecond

        logger.info("Querying NED")
        result_table = Ned.query_region(c, radius=r)

        if len(result_table) == 1:
            if "ZTF" in plot_title:
                plot_title += f' ({result_table["Object Name"][0]})'

            source_coords = [result_table["RA"][0], result_table["DEC"][0]]

            if np.logical_and(
                str(result_table["Redshift"][0]) != "--", source_redshift is None
            ):
                source_redshift = result_table["Redshift"]

            logger.info(
                f"Found likely match to {source_name}"
                f"(type = '{result_table['Type'][0]}'. "
                f"distance = {result_table['Separation'][0]} arcsec')"
            )
        elif len(result_table) > 1:
            logger.warning(
                f"Found multiple possible cross-matches: {result_table['Object Name']}"
            )
        else:
            logger.info("No NED crossmatch found.")

    # Query IRSA, or load from cache

    try:
        os.makedirs(cache_dir)
    except OSError:
        pass

    cache_path = os.path.join(cache_dir, f'{source_name.replace(" ", "")}.csv')

    if from_cache:
        logger.debug(f"Loading from {cache_path}")
        df = pd.read_csv(cache_path)

    else:
        logger.info("Querying IPAC for a lightcurve")
        df = load_irsa(source_coords[0], source_coords[1], radius_arcsec=radius_arcsec)

        logger.debug(f"Saving to {cache_path}")
        if df is not None:
            df.to_csv(cache_path)

    if df is None:
        raise ValueError("No data retrieved from IPAC")
    data = Table.from_pandas(df)

    logger.info(f"There are a total of {len(data)} detections for {source_name}")

    # Start Figure

    plt.figure(figsize=(base_width * 1.1, base_height), dpi=dpi)

    if expanded_labels:
        ax2 = plt.subplot(111)

        ax = ax2.twiny()

    else:
        ax = plt.subplot(111)

    # If you have a redshift, you can add a second y axis!

    if source_redshift is None:
        logger.info("Querying NED to check for a redshift")
        try:
            result_table = Ned.query_object(source_name)
            if len(result_table["Redshift"]) == 1:
                if str(result_table["Redshift"][0]) == "--":
                    raise RemoteServiceError

                source_redshift = result_table["Redshift"][0].data[0]
                logger.info(f"Found a redshift of {source_redshift}")
            elif len(result_table["Redshift"]) > 1:
                logger.warning(f"Found multiple redshifts: {result_table}")
            else:
                raise RemoteServiceError
        except (RemoteServiceError, IndexError) as e:
            logger.info("No redshift found")

    elif np.isnan(source_redshift):
        source_redshift = None

    if source_redshift is not None:
        ax1b = ax.twinx()

        redshift = 1.0 + source_redshift

        if plot_mag:
            dist_mod = 5 * (
                np.log10(cosmo.luminosity_distance(z=source_redshift).to(u.pc).value)
                - 1.0
            )
        else:
            conversion_factor = (
                4
                * np.pi
                * cosmo.luminosity_distance(z=source_redshift).to(u.cm) ** 2.0
                / (redshift)
            )

    cmap = {"zg": "g", "zr": "r", "zi": "orange"}

    wl = {
        "zg": 472.27,
        "zr": 633.96,
        "zi": 788.61,
    }

    markersize = 2.0

    latest_index = list(data["mjd"]).index(max(data["mjd"]))
    latest = data[latest_index]

    dt = format_date(Time(latest["mjd"], format="mjd"), atel=atel)

    logger.info(
        f"Most recent detection on {dt} UT at a magnitude of "
        f"{latest['filtercode'][1]}={latest['mag']:.2f}+/-{latest['magerr']:.2f}"
    )

    # Plot each band (g/r/i)

    for fc in ["zg", "zr", "zi"]:
        mask = data["filtercode"] == fc

        mags = data["mag"][mask] * u.ABmag

        magerrs = (data["magerr"][mask] + data["mag"][mask]) * u.ABmag

        if plot_mag:
            ax.errorbar(
                data["mjd"][mask],
                mags.value,
                yerr=data["magerr"][mask],
                marker="o",
                linestyle=" ",
                markersize=markersize,
                c=cmap[fc],
                label=f"{fc[-1]} ({wl[fc]:.0f} nm)",
            )

            if source_redshift is not None:
                ax1b.errorbar(
                    data["mjd"][mask],
                    mags.value - dist_mod,
                    yerr=data["magerr"][mask],
                    marker="o",
                    linestyle=" ",
                    markersize=markersize,
                    c=cmap[fc],
                    label=f"{fc[-1]} ({wl[fc]:.0f} nm)",
                )

        else:
            flux_j = mags.to(u.Jansky)

            f = (const.c / (wl[fc] * u.nm)).to("Hz")

            flux = (flux_j * f).to("erg cm-2 s-1")

            jerrs = magerrs.to(u.Jansky)
            ferrs = np.abs((jerrs * f).to("erg cm-2 s-1").value - flux.value)

            ax.errorbar(
                data["mjd"][mask],
                flux.to("erg cm-2 s-1").value,
                yerr=ferrs,
                marker="o",
                linestyle=" ",
                markersize=markersize,
                c=cmap[fc],
                label=f"{fc[-1]} ({wl[fc]:.0f} nm)",
            )

            if source_redshift is not None:
                l = flux * conversion_factor

                ax1b.errorbar(
                    data["mjd"][mask],
                    l.to("erg s-1"),
                    marker="o",
                    linestyle=" ",
                    markersize=markersize,
                    c=cmap[fc],
                    label=f"{fc[-1]} ({wl[fc]:.0f} nm)",
                )

    # You can force the y limits if you want

    if ylim is not None:
        ax.set_ylim(ylim)

    if plot_mag:
        ax.set_ylabel(r"Apparent magnitude [AB]", fontsize=big_fontsize)

        ax.invert_yaxis()

        if source_redshift is not None:
            ax1b.set_ylabel(rf"Absolute magnitude [AB]", fontsize=big_fontsize)

            y_min, y_max = ax.get_ylim()

            ax1b.invert_yaxis()

            ax1b.set_ylim(y_min - dist_mod, y_max - dist_mod)

    else:
        ax.set_ylabel(r"$\nu$F$_{\nu}$ [erg cm$^{-2}$ s$^{-1}$]", fontsize=big_fontsize)

        ax.set_yscale("log")

        if source_redshift is not None:
            ax1b.set_ylabel(r"$\nu$L$_{\nu}$ [erg s$^{-1}$]", fontsize=big_fontsize)
            ax1b.set_yscale("log")

            y_min, y_max = ax.get_ylim()
            print(conversion_factor)

            ax1b.set_ylim(
                y_min * conversion_factor.value, y_max * conversion_factor.value
            )

    ax.set_xlabel("Date [MJD]", fontsize=big_fontsize)

    # Add neutrino

    if nu_name is None:
        nu_name = []

    if not isinstance(nu_name, list):
        nu_name = [nu_name]

    for j, nu in enumerate(nu_name):
        gcn_no = find_gcn_no(nu)
        gcn_info = parse_gcn_circular(gcn_no)

        ax.axvline(gcn_info["time"].mjd, linestyle=":", label=nu, color=f"C{j}")

    if expanded_labels:
        # Set up ISO dates

        lmjd, umjd = ax.get_xlim()

        lt = Time(lmjd, format="mjd")
        ut = Time(umjd, format="mjd")

        nt = Time.now()
        nt.format = "fits"

        mjds = []
        labs = []

        for year in range(2016, int(nt.value[:4]) + 1):
            for k, month in enumerate([1, 7]):
                t = Time(f"{year}-{month}-01T00:00:00.01", format="isot", scale="utc")
                t.format = "fits"
                t.out_subfmt = "date"

                if np.logical_and(t > lt, t < ut):
                    mjds.append(t.mjd)
                    labs.append(t.value)

        ax2.set_xticks(mjds)
        ax2.set_xticklabels(labels=labs, rotation=80)

        ax2.set_xlim(lmjd, umjd)

        ax.set_title(f'ZTF Lightcurve of {plot_title.replace("J", " J")}', y=1.4)

        ax2.tick_params(axis="both", which="major", labelsize=big_fontsize)

    ax.tick_params(axis="both", which="both", labelsize=big_fontsize, left=True)

    if source_redshift is not None:
        ax1b.tick_params(axis="both", which="major", labelsize=big_fontsize)

    ax.legend(
        loc="upper center",
        bbox_to_anchor=(0.5, 1.22 + 0.2 * float(expanded_labels)),
        ncol=3 + len(nu_name),
        fancybox=True,
        fontsize=big_fontsize,
    )

    filename = f"{source_name.replace(' ', '')}_lightcurve{['_flux', ''][plot_mag]}.png"

    output_path = os.path.join(plot_folder, f"{filename}")

    logger.info(f"Saving to {output_path}")

    plt.savefig(output_path, bbox_inches="tight", pad_inches=0.05)

    if extra_folder is not None:
        extra_path = os.path.join(extra_folder, f"{filename}")
        logger.info(f"Saving to {extra_path}")
        plt.savefig(extra_path, bbox_inches="tight", pad_inches=0.05)



#!/usr/bin/env python
# coding: utf-8

import json
import logging
import os
import time

import healpy as hp
import matplotlib.pyplot as plt
import numpy as np
import yaml
from astropy.time import Time
from tqdm import tqdm

from nuztf.ampel_api import (
    ampel_api_acknowledge_chunk,
    ampel_api_lightcurve,
    ampel_api_skymap,
    ampel_api_timerange,
    get_preprocessed_results,
)
from nuztf.base_scanner import BaseScanner
from nuztf.paths import BASE_CANDIDATE_DIR, CONFIG_DIR
from nuztf.skymap import Skymap


class SkymapScanner(BaseScanner):
    default_fritz_group = 1563

    def __init__(
        self,
        event: str = None,
        rev: int = None,
        prob_threshold: float = 0.9,
        cone_nside: int = 64,
        output_nside: int | None = None,
        n_days: float = 3.0,  # By default, accept things detected within 72 hours of event time
        config: dict = None,
    ):
        self.logger = logging.getLogger(__name__)
        self.prob_threshold = prob_threshold
        self.n_days = n_days
        self.event = event
        self.prob_threshold = prob_threshold
        self.output_nside = output_nside

        if config:
            self.config = config
        else:
            config_path = CONFIG_DIR.joinpath("gw_run_config.yaml")
            with open(config_path) as f:
                self.config = yaml.safe_load(f)

        self.skymap = Skymap(
            event=self.event,
            rev=rev,
            prob_threshold=self.prob_threshold,
            output_nside=self.output_nside,
        )
        self.rev = self.skymap.rev

        self.t_min = Time(self.skymap.t_obs, format="isot", scale="utc")

        BaseScanner.__init__(
            self,
            run_config=self.config,
            t_min=self.t_min,
            cone_nside=cone_nside,
        )

        self.default_t_max = Time(self.t_min.jd + self.n_days, format="jd")
        self.logger.info(f"Time-range is {self.t_min} -- {self.default_t_max.isot}")

    def get_full_name(self):
        if self.skymap.event_name is not None:
            return self.skymap.event_name
        else:
            return "?????"

    def get_name(self) -> str:
        return f"{self.skymap.event_name}/{self.prob_threshold}"

    def download_results(self):
        """
        Retrieve computed results from the DESY cloud
        """
        self.logger.info("Retrieving results from the DESY cloud")
        file_basename = f"{self.skymap.event}_{self.skymap.rev}"

        res = get_preprocessed_results(file_basename=file_basename)

        if res is None:
            final_objects = []
        else:
            final_objects = [alert["objectId"] for alert in res]
            for alert in res:
                self.cache[alert["objectId"]] = alert

        final_objects = self.remove_duplicates(final_objects)

        self.logger.info(
            f"Retrieved {len(final_objects)} final objects for event "
            f"{self.get_name()} from DESY cloud."
        )

        self.final_candidates = final_objects

    def get_alerts(self):
        """Scan the skymap area and get ZTF transients"""
        self.logger.info("Commencing skymap scan")

        self.logger.debug(
            f"API skymap search: nside = {self.cone_nside} "
            f"/ # pixels = {len(self.cone_ids)} / "
            f"timespan = {self.default_t_max.jd-self.t_min.jd:.1f} days."
        )

        time_healpix_start = time.time()

        self.queue = []

        resume = True
        chunk_size = 2000
        resume_token = None

        i = 0
        total_chunks = 0
        t0 = time.time()

        while resume:
            res, resume_token, chunk_id, remaining_chunks = ampel_api_skymap(
                pixels=self.cone_ids,
                nside=self.cone_nside,
                t_min_jd=self.t_min.jd,
                t_max_jd=self.default_t_max.jd,
                max_n_detections=10,
                logger=self.logger,
                chunk_size=chunk_size,
                resume_token=resume_token,
                warn_exceeding_chunk=False,
            )
            self.queue.extend(res)

            ampel_api_acknowledge_chunk(resume_token=resume_token, chunk_id=chunk_id)

            if i == 0:
                total_chunks = remaining_chunks + 1
                self.logger.info(f"Total chunks: {total_chunks}")

            if remaining_chunks % 50 == 0 and remaining_chunks != 0:
                t1 = time.time()
                processed_chunks = total_chunks - remaining_chunks
                time_per_chunk = (t1 - t0) / processed_chunks
                remaining_time = time_per_chunk * remaining_chunks
                self.logger.info(
                    f"Remaining chunks: {remaining_chunks}. Estimated time to finish: {remaining_time/60:.0f} min"
                )

            if len(res) < chunk_size:
                resume = False
                self.logger.info("Done.")
            else:
                self.logger.debug(
                    f"Chunk size reached ({chunk_size}), commencing next query."
                )
            i += 1

        time_healpix_end = time.time()
        time_healpix = time_healpix_end - time_healpix_start

        cache_file = self.get_cache_dir().joinpath("all_alerts.json")

        with open(cache_file, "w") as outfile:
            json.dump(self.queue, outfile)

        self.n_alerts = len(self.queue)

        self.logger.info(
            f"Added {self.n_alerts} alerts found between {self.t_min}"
            f" and {self.default_t_max.isot}"
        )
        self.logger.info(f"This took {time_healpix:.1f} s in total")

    def filter_alerts(self, load_cachefile=False):
        """ """
        self.logger.info(f"Commencing first stage filtering.")
        cache_file = self.get_cache_dir().joinpath("all_alerts.json")

        if load_cachefile:
            self.queue = json.load(open(cache_file, "r"))

        first_stage_objects = []
        filter_time_start = time.time()

        i_survived = []

        for i, res in enumerate(tqdm(self.queue)):
            ztf_id = res["objectId"]

            if self.filter_f_no_prv(
                res=res,
            ):
                self.logger.debug(
                    f"{ztf_id}: Passed first cut (does not have previous detections)."
                )
                if self.filter_ampel(res):
                    self.logger.debug(f"{ztf_id}: Passed AMPEL cut.")
                    i_survived.append(i)
                else:
                    self.logger.debug(f"{ztf_id}: Failed AMPEL cut.")
            else:
                self.logger.debug(
                    f"{ztf_id}: Failed first cut (has previous detections)."
                )

        first_stage_objects = [self.queue[i]["objectId"] for i in i_survived]
        first_stage_objects = self.remove_duplicates(first_stage_objects)

        filter_time_end = time.time()
        filter_time = filter_time_end - filter_time_start

        self.logger.info(
            f"First stage of filtering (based on predetections plus AMPEL cuts) "
            f"took {filter_time:.1f} s in total. "
            f"{len(first_stage_objects)} transients make the cut."
        )

        cache_file_first_stage = self.get_cache_dir().joinpath("first_stage.json")

        with open(cache_file_first_stage, "w") as outfile:
            json.dump(first_stage_objects, outfile)

        # Second and final stage
        self.logger.info(
            f"Second stage commencing: Now we do additional filtering based on history."
        )

        start_secondfilter = time.time()

        final_objects = []

        for ztf_id in tqdm(first_stage_objects):
            # Get the full lightcurve from the API
            query_res = ampel_api_lightcurve(ztf_name=ztf_id, logger=self.logger)

            for res in query_res:
                _ztf_id = res["objectId"]

                if self.filter_f_history(res=res):
                    final_objects.append(_ztf_id)
                    self.cache[_ztf_id] = res
                    self.logger.debug(f"✅ {_ztf_id}: Passed all filters.")
                else:
                    self.logger.debug(f"❌ {_ztf_id}: Failed History.")

        end_secondfilter = time.time()
        filter_time = end_secondfilter - start_secondfilter

        final_objects = self.remove_duplicates(final_objects)

        cache_file_final_stage = self.get_cache_dir().joinpath("final_stage.json")

        with open(cache_file_final_stage, "w") as outfile:
            json.dump(final_objects, outfile)

        self.logger.info(
            f"Final stage of filtering took {filter_time:.1f} s in total. "
            f"{len(final_objects)} transients make the cut."
        )

        self.final_candidates = final_objects

    def remove_duplicates(self, ztf_ids: list):
        """ """
        return list(set(ztf_ids))

    def get_obs_line(self):
        """ """
        return "Each exposure was 30s with a typical depth of 20.5 mag."

    @staticmethod
    def remove_variability_line():
        """ """
        return (
            ", and removing candidates with history of "
            "variability prior to the merger time"
        )

    def candidate_text(
        self, ztf_id: str, first_detection: float, lul_lim: float, lul_jd: float
    ):
        """ """
        try:
            text = (
                "{0}, first detected {1:.1f} hours after merger, "
                "was not detected {2:.1f} days prior to a depth of {3:.2f}. ".format(
                    ztf_id,
                    24.0 * (first_detection - self.t_min.jd),
                    first_detection - lul_jd,
                    lul_lim,
                )
            )
        except TypeError:
            text = (
                f"{ztf_id} had upper limit problems. PLEASE FILL IN NUMBERS BY HAND!!! "
            )

        return text

    def filter_f_no_prv(self, res: dict, t_max_jd=None) -> bool:
        """First filtering stage"""

        if t_max_jd is None:
            t_max_jd = self.default_t_max.jd

        # Veto transients older than t_min_jd
        # as we don't expect detections before GRB or GW event time)
        if res["candidate"]["jdstarthist"] < self.t_min.jd:
            startdate_jd = res["candidate"]["jdstarthist"]
            startdate_date = Time(startdate_jd, format="jd").isot
            self.logger.debug(
                f"❌ {res['objectId']}: Transient is too old "
                f"(jdstarthist predates event; first detection at {startdate_date})."
            )
            return False

        # Veto new transients
        if res["candidate"]["jdstarthist"] > t_max_jd:
            startdate_jd = res["candidate"]["jdstarthist"]
            startdate_date = Time(startdate_jd, format="jd").isot
            self.logger.debug(
                f"❌ {res['objectId']}: Transient is too new "
                f"(jdstarthist too late after event; "
                f"first detection at {startdate_date}, "
                f"filter searches only up to {Time(t_max_jd, format='jd').isot})."
            )
            return False

        # Exclude negative detection
        if res["candidate"]["isdiffpos"] not in ["t", "1"]:
            self.logger.debug(f"❌ {res['objectId']}: Negative subtraction")
            return False

        try:
            if res["candidate"]["drb"] < 0.3:
                self.logger.debug(f"❌ {res['objectId']}: DRB too low")
                return False
        except (KeyError, TypeError):
            pass

        # Check contour
        if not self.skymap.in_contour(res["candidate"]["ra"], res["candidate"]["dec"]):
            self.logger.debug(f"❌ {res['objectId']}: Outside of event contour.")
            return False

        self.logger.debug(
            f"✅ {res['objectId']}: Passes first filtering stage (no prv)"
        )

        return True

    def filter_f_history(self, res: dict, t_max_jd=None):
        """Veto transients"""

        if t_max_jd is None:
            t_max_jd = self.default_t_max.jd

        # Veto old transients
        ztf_id = res["objectId"]
        if res["candidate"]["jdstarthist"] < self.t_min.jd:
            self.logger.debug(
                f"❌ {ztf_id}: Transient is too old. (jdstarthist history predates event)"
            )
            return False

        # Veto new transients
        if res["candidate"]["jdstarthist"] > t_max_jd:
            self.logger.debug(
                f"❌ {ztf_id}: Transient is too new. (jdstarthist too late after event)"
            )
            return False

        # Require 2 detections separated by 15 mins
        if (res["candidate"]["jdendhist"] - res["candidate"]["jdstarthist"]) < 0.01:
            self.logger.debug(f"❌ {ztf_id}: Not passed mover cut")
            return False

        # Require 2 positive detections
        old_detections = [
            x
            for x in res["prv_candidates"]
            if np.logical_and("isdiffpos" in x.keys(), x["jd"] > self.t_min.jd)
        ]

        pos_detections = [x for x in old_detections if "isdiffpos" in x.keys()]

        if len(pos_detections) < 1:
            self.logger.debug(f"❌ {ztf_id}: Does not have two detections")
            return False

        self.logger.debug(f"✅ {ztf_id}: Passed the history filtering stage")

        return True

    def unpack_skymap(self, output_nside: None | int = None):
        """ """
        if output_nside is not None:
            self.skymap = Skymap(
                event=self.event,
                rev=self.rev,
                prob_threshold=self.prob_threshold,
                output_nside=output_nside,
            )

        nside = hp.npix2nside(len(self.skymap.data[self.skymap.key]))

        mask = self.skymap.data[self.skymap.key] > self.skymap.pixel_threshold

        map_coords = []

        pixel_nos = []

        self.logger.info("Checking which pixels are within the contour:")

        for i in tqdm(range(hp.nside2npix(nside))):
            if mask[i]:
                map_coords.append(self.extract_ra_dec(nside, i))
                pixel_nos.append(i)

        total_pixel_area = hp.nside2pixarea(nside, degrees=True) * float(
            len(map_coords)
        )

        self.logger.info(f"Total pixel area: {total_pixel_area} degrees")

        map_coords = np.array(
            map_coords, dtype=np.dtype([("ra", float), ("dec", float)])
        )

        return (
            map_coords,
            pixel_nos,
            nside,
            self.skymap.data[self.skymap.key][mask],
            self.skymap.data,
            total_pixel_area,
            self.skymap.key,
        )

    def find_cone_coords(self):
        """ """

        cone_ids = []

        for ra, dec in self.map_coords:
            cone_ids.append(self.extract_npix(self.cone_nside, ra, dec))

        cone_ids = list(set(cone_ids))

        cone_coords = []

        for i in tqdm(cone_ids):
            cone_coords.append(self.extract_ra_dec(self.cone_nside, i))

        cone_coords = np.array(
            cone_coords, dtype=np.dtype([("ra", float), ("dec", float)])
        )

        return cone_ids, cone_coords

    def plot_skymap(self):
        """ """
        fig = plt.figure()
        plt.subplot(111, projection="aitoff")

        mask = self.data[self.key] > self.skymap.pixel_threshold

        size = hp.max_pixrad(self.nside, degrees=True) ** 2

        ra_map_rad = np.deg2rad(self.wrap_around_180(self.map_coords["ra"]))
        dec_map_rad = np.deg2rad(self.map_coords["dec"])

        plt.scatter(
            ra_map_rad,
            dec_map_rad,
            c=self.data[self.key][mask],
            vmin=0.0,
            vmax=max(self.data[self.key]),
            s=size,
        )

        plt.title("SKYMAP")

        outpath = self.get_output_dir().joinpath("skymap.png")
        plt.tight_layout()

        plt.savefig(outpath, dpi=300)

        return fig

    def plot_coverage(self, plot_candidates: bool = True):
        """Plot ZTF coverage of skymap region"""
        fig, message = self.plot_overlap_with_observations(
            first_det_window_days=self.n_days
        )

        if plot_candidates:
            for candidate, res in self.cache.items():
                ra = np.deg2rad(
                    self.wrap_around_180(np.array([res["candidate"]["ra"]]))
                )
                dec = np.deg2rad(res["candidate"]["dec"])

                plt.scatter(
                    ra, dec, color="white", marker="*", s=50.0, edgecolor="black"
                )

        plt.tight_layout()

        outpath = self.get_output_dir().joinpath("coverage.png")

        plt.savefig(outpath, dpi=300)

        self.logger.info(message)

        return fig, message



import os

import backoff
import requests  # type: ignore

from nuztf.credentials import load_credentials

# Fritz API URLs

API_BASEURL = "https://fritz.science"

fritz_token = load_credentials("fritz", token_based=True)


def fritz_api(method: str, endpoint_extension: str, data: dict = None):
    headers = {"Authorization": f"token {fritz_token}"}
    endpoint = os.path.join(API_BASEURL, endpoint_extension)
    if method in ["post", "POST"]:
        response = requests.request(method, endpoint, json=data, headers=headers)
    elif method in ["get", "GET"]:
        response = requests.request(method, endpoint, params=data, headers=headers)
    else:
        raise ValueError("You have to use either 'get' or 'post'")
    return response


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=60,
)
def save_source_to_group(object_id: str, group_id: int):
    payload = {
        "objId": object_id,
        "inviteGroupIds": [group_id],
    }
    return fritz_api(
        method="POST", endpoint_extension="api/source_groups", data=payload
    )


def delete_source_from_group(object_id: str, group_id: int):
    payload = {
        "objId": object_id,
        "unsaveGroupIds": [group_id],
    }
    return fritz_api(
        method="POST", endpoint_extension="api/source_groups", data=payload
    )



#!/usr/bin/env python3
# coding: utf-8

import logging
import os
import subprocess

import matplotlib.pyplot as plt
import seaborn as sns
from ztfquery.io import LOCALSOURCE

logger = logging.getLogger(__name__)

sns.set_style("white")

# Use latex if available
try:
    subprocess.check_output(["which", "latex"])
    plt.rc("text", usetex=True)
    plt.rc("text.latex", preamble=r"\usepackage{romanbar}")
except subprocess.CalledProcessError:
    logger.warning(
        "No Latex installation found. Proceeding without, but plots may look weird."
    )

plt.rcParams["font.family"] = "sans-serif"

plot_dir = os.path.join(LOCALSOURCE, "plots")
if not os.path.exists(plot_dir):
    os.makedirs(plot_dir)

dpi = 300

fontsize = 7.0
big_fontsize = 10.0

golden_ratio = 1.618

base_width = 4.0
base_height = base_width / golden_ratio

margin_width = 0.5 * base_width
margin_height = margin_width / golden_ratio

full_width = 1.5 * base_width
full_height_landscape = full_width / golden_ratio
full_height_a4 = 11.75 / 8.25 * full_width

cmap = "rocket"



#!/usr/bin/env python3

import gzip
import io
import logging
from base64 import b64encode
from json import JSONDecodeError
from pathlib import Path

import backoff
import numpy as np
import requests
from ampel.util.json import load
from ampel.ztf.util.ZTFIdMapper import ZTFIdMapper
from astropy.io import fits  # type: ignore
from astropy.time import Time  # type: ignore
from requests.auth import HTTPBasicAuth
from requests.exceptions import HTTPError

from nuztf import utils
from nuztf.credentials import load_credentials
from nuztf.paths import PREPROCESSED_CACHE_DIR

API_BASEURL = "https://ampel.zeuthen.desy.de"
API_ZTF_ARCHIVE_URL = API_BASEURL + "/api/ztf/archive/v3"
API_CATALOGMATCH_URL = API_BASEURL + "/api/catalogmatch"

MAX_N_PIX = 1000

_, ampel_api_archive_token = load_credentials("ampel_api_archive_token")


def merge_alerts(alert_list: list) -> list:
    """ """
    merged_list = []
    keys = list(set([x["objectId"] for x in alert_list]))

    for objectid in keys:
        alerts = [x for x in alert_list if x["objectId"] == objectid]
        if len(alerts) == 1:
            merged_list.append(alerts[0])
        else:
            jds = [x["candidate"]["jd"] for x in alerts]
            order = [jds.index(x) for x in sorted(jds)[::-1]]
            latest = alerts[jds.index(max(jds))]
            latest["candidate"]["jdstarthist"] = min(
                [x["candidate"]["jdstarthist"] for x in alerts]
            )

            for index in order[1:]:
                x = alerts[index]

                # Merge previous detections

                for prv in x["prv_candidates"] + [x["candidate"]]:
                    if prv not in latest["prv_candidates"]:
                        latest["prv_candidates"] = [prv] + latest["prv_candidates"]

            merged_list.append(latest)

    return merged_list


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=600,
)
def ampel_api_cone(
    ra: float,
    dec: float,
    radius: float,
    t_min_jd=Time("2018-04-01T00:00:00.123456789", format="isot", scale="utc").jd,
    t_max_jd=Time.now().jd,
    with_history: bool = False,
    with_cutouts: bool = False,
    chunk_size: int = 500,
    logger=None,
) -> list:
    """Query ampel via a cone search"""

    if logger is None:
        logger = logging.getLogger(__name__)

    if with_history:
        hist = "true"
    else:
        hist = "false"

    if with_cutouts:
        cutouts = "true"
    else:
        cutouts = "false"

    queryurl_conesearch = (
        API_ZTF_ARCHIVE_URL + f"/alerts/cone_search?ra={ra}&dec={dec}&"
        f"radius={radius}&jd_start={t_min_jd}&"
        f"jd_end={t_max_jd}&with_history={hist}&"
        f"with_cutouts={cutouts}&chunk_size={chunk_size}"
    )

    logger.debug(queryurl_conesearch)

    headers = {"Authorization": f"Bearer {ampel_api_archive_token}"}

    response = requests.get(
        queryurl_conesearch,
        headers=headers,
    )

    if response.status_code == 503:
        raise requests.exceptions.RequestException

    try:
        query_res = [i for i in response.json()["alerts"]]
    except JSONDecodeError:
        if response.headers:
            logger.debug(response.headers)
        raise requests.exceptions.RequestException

    nr_results = len(query_res)

    logger.debug(f"Found {nr_results} alerts.")

    if nr_results == chunk_size:
        logger.warning(
            f"Query result limited by chunk size! You will most likely be missing alerts!"
        )

    return query_res


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=600,
)
def ampel_api_timerange(
    t_min_jd=Time("2018-04-01T00:00:00.123456789", format="isot", scale="utc").jd,
    t_max_jd=Time.now().jd,
    with_history: bool = False,
    with_cutouts: bool = False,
    chunk_size: int = 500,
    logger=None,
) -> list:
    """Function to query ampel via a time-range search"""

    if logger is None:
        logger = logging.getLogger(__name__)

    if with_history:
        hist = "true"
    else:
        hist = "false"

    if with_cutouts:
        cutouts = "true"
    else:
        cutouts = "false"

    queryurl_timerange = (
        API_ZTF_ARCHIVE_URL + f"/alerts/time_range?jd_start={t_min_jd}&"
        f"jd_end={t_max_jd}&with_history={hist}&"
        f"with_cutouts={cutouts}&chunk_size={chunk_size}"
    )

    logger.debug(queryurl_timerange)

    headers = {"Authorization": f"Bearer {ampel_api_archive_token}"}

    response = requests.get(
        queryurl_timerange,
        headers=headers,
    )

    if response.status_code == 503:
        raise requests.exceptions.RequestException

    try:
        query_res = [i for i in response.json()["alerts"]]
    except JSONDecodeError:
        if response.headers:
            logger.debug(response.headers)
        raise requests.exceptions.RequestException

    nr_results = len(query_res)

    logger.debug(f"Found {nr_results} alerts.")

    if nr_results == chunk_size:
        logger.warning(
            f"Query result limited by chunk size! You will most likely be missing alerts!"
        )

    return query_res


def ensure_cutouts(alert: list, logger=None):
    """Make sure alert contains cutouts (if not, query them from AMPEL API"""

    if logger is None:
        logger = logging.getLogger(__name__)

    candid = alert[0]["candid"]
    ztf_id = alert[0]["objectId"]

    if "cutoutScience" in alert[0].keys():
        if "stampData" in alert[0]["cutoutScience"].keys():
            logger.debug("Alert already contains cutouts.")

            return alert

    logger.debug(f"{ztf_id}: Querying API for cutouts.")

    final_cutouts = {}

    cutouts = ampel_api_cutout(candid)

    if "detail" in cutouts.keys():
        if cutouts["detail"] == "Not Found":
            for k in [
                "science",
                "difference",
                "template",
                "Cutoutscience",
                "Cutoutdifference",
                "Cutouttemplate",
            ]:
                final_cutouts[f"cutout{k.title()}"] = {
                    "stampData": create_empty_cutout()
                }
    else:
        for k in cutouts:
            final_cutouts[f"cutout{k.title()}"] = {"stampData": cutouts[k]}

    alert[0] = {**alert[0], **final_cutouts}

    logger.debug(f"{ztf_id}: Added cutouts.")

    return alert


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=600,
)
def ampel_api_name(
    ztf_name: str,
    with_history: bool = True,
    with_cutouts: bool = False,
    limit: int = 999999,
    logger=None,
) -> list:
    """Function to query ampel via name"""
    if logger is None:
        logger = logging.getLogger(__name__)

    if with_history:
        hist = "true"
    else:
        hist = "false"

    if with_cutouts:
        cutouts = "true"
    else:
        cutouts = "false"

    queryurl_ztf_name = (
        API_ZTF_ARCHIVE_URL
        + f"/object/{ztf_name}/alerts?with_history={hist}&with_cutouts={cutouts}&limit={limit}"
    )

    logger.debug(queryurl_ztf_name)

    headers = {"Authorization": f"Bearer {ampel_api_archive_token}"}

    response = requests.get(
        queryurl_ztf_name,
        headers=headers,
    )

    if response.status_code == 503:
        raise requests.exceptions.RequestException

    try:
        query_res = [i for i in response.json()]
        query_res = merge_alerts(query_res)

    except JSONDecodeError:
        if response.headers:
            logger.debug(response.headers)
        raise requests.exceptions.RequestException

    return query_res


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=600,
)
def ampel_api_lightcurve(
    ztf_name: str,
    t_min_jd=Time("2017-01-01T00:00:00.0", format="isot", scale="utc").jd,
    t_max_jd=Time.now().jd,
    program_id: int = None,
    logger=None,
) -> list:
    """
    Function to query ampel via name, returns a virtual alert
    constructed by AMPEL containing ALL photopoints and upper limits

    """

    if logger is None:
        logger = logging.getLogger(__name__)

    if program_id is None:
        queryurl_lightcurve = (
            API_ZTF_ARCHIVE_URL + f"/object/{ztf_name}/photopoints?jd_start={t_min_jd}&"
            f"jd_end={t_max_jd}"
        )
    else:
        queryurl_lightcurve = (
            API_ZTF_ARCHIVE_URL + f"/object/{ztf_name}/photopoints?jd_start={t_min_jd}&"
            f"jd_end={t_max_jd}&programid={program_id}"
        )

    logger.debug(queryurl_lightcurve)

    headers = {"Authorization": f"Bearer {ampel_api_archive_token}"}

    response = requests.get(
        queryurl_lightcurve,
        headers=headers,
    )

    if response.status_code == 503:
        if response.headers:
            logger.debug(response.headers)
        raise requests.exceptions.RequestException

    try:
        query_res = [response.json()]

    except JSONDecodeError:
        if response.headers:
            logger.debug(response.headers)
        raise requests.exceptions.RequestException

    return query_res


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=600,
)
def ampel_api_healpix(
    ipix: int,
    nside: int = 64,
    t_min_jd=Time("2018-04-01T00:00:00.123456789", format="isot", scale="utc").jd,
    t_max_jd=Time.now().jd,
    with_history: bool = False,
    with_cutouts: bool = False,
    chunk_size: int = 500,
    logger=None,
) -> list:
    """Function to query ampel based on a healpix pixel-index (nside is the pixelization degree)"""

    if logger is None:
        logger = logging.getLogger(__name__)

    if with_history:
        hist = "true"
    else:
        hist = "false"

    if with_cutouts:
        cutouts = "true"
    else:
        cutouts = "false"

    queryurl_healpix = (
        API_ZTF_ARCHIVE_URL
        + f"/alerts/healpix?nside={nside}&ipix={ipix}&jd_start={t_min_jd}&jd_end={t_max_jd}&with_history={hist}&with_cutouts={cutouts}&chunk_size={chunk_size}"
    )

    logger.debug(queryurl_healpix)

    headers = {"Authorization": f"Bearer {ampel_api_archive_token}"}

    response = requests.get(
        queryurl_healpix,
        headers=headers,
    )

    if response.status_code == 503:
        raise requests.exceptions.RequestException

    try:
        query_res = [i for i in response.json()["alerts"]]
    except JSONDecodeError:
        if response.headers:
            logger.debug(response.headers)
        raise requests.exceptions.RequestException

    nr_results = len(query_res)

    logger.debug(f"Found {nr_results} alerts.")

    if nr_results == chunk_size:
        logger.warning(
            f"Query result limited by chunk size! You will most likely be missing alerts!"
        )

    return query_res


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=1200,
)
def ampel_api_acknowledge_chunk(resume_token: str, chunk_id: int, logger=None):
    """
    After receiving a chunk, acknowledge that we got it
    (otherwise large alert queries will start looping)
    """
    if logger is None:
        logger = logging.getLogger(__name__)

    endpoint = (
        API_ZTF_ARCHIVE_URL + f"/stream/{resume_token}/chunk/{chunk_id}/acknowledge"
    )

    headers = {
        "accept": "application/json",
        "Content-Type": "application/json",
        "Authorization": f"Bearer {ampel_api_archive_token}",
    }

    payload = {"resume_token": resume_token, "chunk_id": chunk_id}

    logger.debug(f"Acknowledging:\n{payload}")

    response = requests.post(url=endpoint, json=payload, headers=headers)


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=1200,
)
def ampel_api_skymap(
    pixels: list,
    nside: int = 64,
    t_min_jd=Time("2018-04-01T00:00:00.123456789", format="isot", scale="utc").jd,
    t_max_jd=Time.now().jd,
    max_n_detections: int = 99999999,
    with_history: bool = False,
    with_cutouts: bool = False,
    chunk_size: int = 500,
    resume_token: str = None,
    warn_exceeding_chunk: bool = True,
    program_id: int = None,
    logger=None,
) -> tuple:
    """
    Function to query ampel based on lists of healpix pixels of different resolution (nside is the respective resolution)
    """

    if logger is None:
        logger = logging.getLogger(__name__)

    queryurl_skymap = API_ZTF_ARCHIVE_URL + f"/alerts/healpix/skymap"

    # First, we create a json body to post
    headers = {
        "accept": "application/json",
        "Content-Type": "application/json",
        "Authorization": f"Bearer {ampel_api_archive_token}",
    }
    if with_history:
        hist = "true"
    else:
        hist = "false"

    if with_cutouts:
        cutouts = "true"
    else:
        cutouts = "false"

    # if we have a resume_token to proceed to the next chunk, that's all we need
    if resume_token is not None:
        queryurl_stream = API_ZTF_ARCHIVE_URL + f"/stream/{resume_token}/chunk"
        response = requests.get(
            queryurl_stream, params={"with_history": hist}, headers=headers
        )

    # if we don't have a resume_token, we first need to create the full query
    else:
        # Now we reduce the query size
        regions = utils.deres(nside=nside, ipix=pixels)

        n_pix = 0
        for reg in regions:
            n_pix += len(reg["pixels"])

        logger.debug(f"This comprises {n_pix} individual pixels")

        if n_pix > MAX_N_PIX:
            logger.warning(
                f"Total number of pixels exceeds threshold ({MAX_N_PIX} pixels). Issuing a query for the full sky instead."
            )
            regions = [
                {"nside": 1, "pixels": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]}
            ]

        query = {
            "regions": regions,
            "jd": {
                "$lt": t_max_jd,
                "$gt": t_min_jd,
            },
            "latest": "false",
            "candidate": {
                "rb": {"$gt": 0.3},
                "magpsf": {"$gt": 15},
                "ndethist": {"$gt": 0, "$lte": max_n_detections},
            },
            "with_history": hist,
            "with_cutouts": cutouts,
            "chunk_size": chunk_size,
        }

        if resume_token:
            query["resume_token"] = resume_token

        if program_id is not None:
            query["programid"] = program_id

        logger.debug(f"Query url:\n{queryurl_skymap}")
        logger.debug(f"Query:\n{query}")

        response = requests.post(url=queryurl_skymap, json=query, headers=headers)

    logger.debug(response)
    logger.debug(response.status_code)

    if response.status_code in [424, 503]:
        raise requests.exceptions.RequestException

    try:
        res_json = response.json()
        remaining_chunks = res_json["remaining"]["chunks"]
        logger.debug(f"Remaining chunks: {remaining_chunks}")
        chunk_id = res_json.get("chunk", None)
        resume_token = response.json().get("resume_token", None)
        query_res = [i for i in response.json()["alerts"]]
    except JSONDecodeError:
        if response.headers:
            logger.debug(response.headers)
        raise requests.exceptions.RequestException

    nr_results = len(query_res)

    logger.debug(f"Found {nr_results} alerts.")

    if nr_results == chunk_size and warn_exceeding_chunk:
        logger.warning(
            f"Query result limited by chunk size! You will most likely be missing alerts!"
        )

    return query_res, resume_token, chunk_id, remaining_chunks


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=600,
)
def ampel_api_cutout(candid: int, logger=None):
    """Function to query ampel for cutouts by candidate ID"""

    if logger is None:
        logger = logging.getLogger(__name__)

    if "v2" in API_ZTF_ARCHIVE_URL:
        queryurl_cutouts = API_ZTF_ARCHIVE_URL + f"/cutouts/{candid}"
    else:
        queryurl_cutouts = API_ZTF_ARCHIVE_URL + f"/alert/{candid}/cutouts"

    headers = {"Authorization": f"Bearer {ampel_api_archive_token}"}

    response = requests.get(
        queryurl_cutouts,
        headers=headers,
    )

    logger.debug(queryurl_cutouts)

    if response.status_code == 503:
        raise requests.exceptions.RequestException

    try:
        cutouts = response.json()
    except JSONDecodeError:
        if response.headers:
            logger.debug(response.headers)
        raise requests.exceptions.RequestException

    return cutouts


def create_empty_cutout():
    """Function to reate an empty image for missing cutouts"""
    npix = 63

    blank = np.ones((npix, npix))

    for i in range(npix):
        c = abs(npix / 2 - i) / (0.5 * npix)
        blank[i - 1][i - 1] = c
        blank[i - 1][npix - i - 1] = c

    hdu = fits.PrimaryHDU(blank)
    hdul = fits.HDUList([hdu])
    comp = io.BytesIO()
    hdul.writeto(comp)
    blank_compressed = gzip.compress(comp.getvalue())
    blank_compressed = b64encode(blank_compressed)

    return blank_compressed


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=600,
)
def ampel_api_catalog(
    catalog: str,
    catalog_type: str,
    ra_deg: float,
    dec_deg: float,
    search_radius_arcsec: float = 10,
    search_type: str = "all",
    logger=None,
):
    """
    Method for querying catalogs via the Ampel API
    'catalog' must be the name of a supported catalog, e.g.
    SDSS_spec, PS1, NEDz_extcats...
    For a full list of catalogs, confer
    https://ampel.zeuthen.desy.de/api/catalogmatch/catalogs

    """
    if not catalog_type in ["extcats", "catsHTM"]:
        raise ValueError(
            f"Expected parameter catalog_type in ['extcats', 'catsHTM'], got {catalog_type}"
        )
    if not search_type in ["all", "nearest"]:
        raise ValueError(
            f"Expected parameter catalog_type in ['all', 'nearest'], got {search_type}"
        )

    if logger is None:
        logger = logging.getLogger(__name__)

    queryurl_catalogmatch = API_CATALOGMATCH_URL + "/cone_search/" + search_type

    # First, we create a json body to post
    headers = {"accept": "application/json", "Content-Type": "application/json"}
    query = {
        "ra_deg": ra_deg,
        "dec_deg": dec_deg,
        "catalogs": [
            {"name": catalog, "rs_arcsec": search_radius_arcsec, "use": catalog_type}
        ],
    }

    logger.debug(queryurl_catalogmatch)
    logger.debug(query)

    response = requests.post(url=queryurl_catalogmatch, json=query, headers=headers)

    if response.status_code == 503:
        if response.headers:
            logger.debug(response.headers)
        raise requests.exceptions.RequestException

    try:
        res = response.json()[0]
    except JSONDecodeError:
        if response.headers:
            logger.debug(response.headers)
        raise requests.exceptions.RequestException

    return res


def get_preprocessed_results(file_basename: str, logger=None) -> None | list:
    """
    Access the DESY Cloud to look if there are precomputed results from an AMPEL run
    there
    """
    if logger is None:
        logger = logging.getLogger(__name__)

    desy_cloud_token = load_credentials("desy_cloud_token", token_based=True)

    filename = PREPROCESSED_CACHE_DIR.joinpath(f"{file_basename}.json.gz")

    res = requests.get(
        f"https://syncandshare.desy.de/public.php/webdav/{filename.name}",
        headers={"X-Requested-With": "XMLHttpRequest"},
        auth=(desy_cloud_token, "bla"),
    )

    if res.status_code != 200:
        logger.warning(
            "\n\n-------------------- !! -------------------\nSomething went wrong with your query.\nCheck your credentials and make sure Ampel\nhas run correctly at Desy.\n-------------------- !! -------------------\n\n"
        )
        return None

    with open(f"{filename}", "wb") as f:
        f.write(res.content)

    res = []
    with gzip.open(filename, "rb") as f_in:
        data = load(f_in)
        for t in data:
            ztf_id = ZTFIdMapper.to_ext_id(t.stock.get("stock"))
            pp = t.get_photopoints()
            pp_reformatted = utils.reformat_downloaded_results(
                photopoints=pp, ztf_id=ztf_id
            )
            redshifts = t.get_latest_t2_body(unit="T2DigestRedshifts")
            kilonova_eval = t.get_latest_t2_body(unit="T2KilonovaEval")
            pp_reformatted.update({"kilonova_eval": kilonova_eval})
            pp_reformatted.update({"redshifts": redshifts})
            res.append(pp_reformatted)

    return res
